from os.path import isfile, abspath

from nio.types.base import Type


class FileHolder(object):
    """ Instance received when assigning this property to a variable """

    def __init__(self, value, mode):
        # actual property value
        self._value = value
        # stream obtained from __enter__ method
        self._stream = None
        # file mode to use when opening file
        self._mode = mode

    def __enter__(self):
        """ Return a file stream when a a valid file exists """
        real_file = self.file
        if real_file is not None:
            # specify mode only when it is user defined, otherwise
            # allow to pick up default value
            self._stream = open(real_file, mode=self._mode) \
                if self._mode else open(real_file)
            return self._stream

    def __exit__(self, _type, value, tb):
        """ Close file stream """
        if self._stream:
            self._stream.close()
            self._stream = None

    # Actual property value
    @property
    def value(self):
        return self._value

    @property
    def file(self):
        # First, just see if it is already a file?
        if isfile(self.value):
            return self.value

        # Next, try the absolute file path
        filename = abspath(self.value)
        if isfile(filename):
            return filename

        return None


class FileType(Type):

    @staticmethod
    def serialize(value, **kwargs):
        """ Convert a value to a JSON serializable value """
        return value

    @staticmethod
    def deserialize(value, **kwargs):
        """ Convert value to file """
        try:
            mode = kwargs.get('mode', None)
            return FileHolder(value, mode)
        except:
            raise TypeError(
                "Unable to cast value to a file: {}".format(
                    value, **kwargs))



from importlib import import_module
from nio.util.logging import get_nio_logger
from nio.util.versioning.check import compare_versions, VersionCheckResult


class DependsOn(object):

    """Decorator for setting a component version dependency on a component."""

    attribute_format = "{0}_dependencies"

    def __init__(self, dependency, version=None):
        """Defines the component version dependency on the decorated component.

        Args:
            dependency (string): the module that is a dependency of the
                                 decorated component.
            version (string): the required version of the dependency.

        """

        self.logger = get_nio_logger("Dependency")
        self._dependency = dependency
        self._version = version

    def __call__(self, cls):
        component_version = None

        # bother with version at all?
        if self._version:
            m = import_module(self._dependency)
            if hasattr(m, "__version__"):
                component_version = m.__version__
                if not validate_version(component_version, self._version):
                    self.logger.warning(
                        "'{}' dependency: '{}' not satisfied, version: '{}' "
                        "is installed and '{}' was requested".format(
                            cls.__name__, self._dependency, component_version,
                            self._version))
            else:
                self.logger.warning('Component: {0} does not contain '
                                    'version info'.
                                    format(self._dependency))

        # add to list of dependencies
        attribute = DependsOn.attribute_format.format(cls.__name__)
        if not hasattr(cls, attribute):
            setattr(cls, attribute, {})
        dependencies = getattr(cls, attribute)
        dependencies[self._dependency] = (component_version, self._version)

        return cls


def validate_version(component_version, dependency_version):
    """Check if a component's version is valid for a given dependency.

    Args:
        component_version (string): version of installed component.
        dependency_version (string): version required by dependency.

    Returns:
        bool: Returns True if valid, False if not.

    """

    # Any version set to None is considered valid for comparison
    # purposes once it gets this far
    if component_version is None or dependency_version is None:
        return True

    return compare_versions(component_version,
                            dependency_version).value > \
        VersionCheckResult.older.value


def get_class_dependencies(cls):
    """Provides class dependencies.

    When parent classes have dependencies these are added simulating
    inheritance behaviour

    Args:
        cls (class type): the class.

    Returns:
        class dependencies defined using @DependsOn

    """
    attribute = DependsOn.attribute_format.format(cls.__name__)
    dependencies = {}
    for _class in cls.__bases__:
        dependencies.update(get_class_dependencies(_class))
    dependencies.update(getattr(cls, attribute, {}))

    return dependencies



from datetime import datetime


def get_nio_time():
    """ Generate a nicely formatted ISO 8601 time string for a log entry

    Returns:
        ISO 8601-conformant time string
    """
    return datetime.utcnow().isoformat() + "Z"



from nio.router.base import BlockRouter
from nio.util.threading import spawn


class ThreadedBlockRouter(BlockRouter):

    """A router that delivers signals asynchronously"""

    def deliver_signals(self, block_receiver, signals):
        """ Delivers signals to given block in a new thread """
        spawn(self.notify_signals_to_block, block_receiver, signals)



import logging
from nio.util.logging.adapter import NIOAdapter


class LoggingHelper(object):

    _prefix = None

    @classmethod
    def get_nio_logger(cls, logger_name=None):
        if not logger_name:
            logger_name = 'default'

        if cls._prefix:
            logger_name = "{}.{}".format(cls._prefix, logger_name)

        return NIOAdapter(logging.getLogger(logger_name))

    @classmethod
    def _register_prefix(cls, prefix_name):
        if prefix_name:
            cls._prefix = prefix_name
        else:
            cls._deregister_prefix()

    @classmethod
    def _deregister_prefix(cls):
        cls._prefix = None



import inspect
from nio.properties import BaseProperty
from nio.properties import PropertyHolder
from nio.properties.util.object_type import ObjectType


class ObjectProperty(BaseProperty):
    """ Defines a property for an object type.

    Object types contain properties themselves, and must inherit from
    PropertyHolder just like the parent class.

    """

    def __init__(self, obj_type, **kwargs):
        """ Initializes the property.

        Args:
            obj_type (class): class type which is an instance of PropertyHolder
        """
        # Validate that the object is a PropertyHolder
        if not inspect.isclass(obj_type) or \
                not issubclass(obj_type, PropertyHolder):
            raise TypeError(
                "Specified object type {} is not a PropertyHolder".format(
                    obj_type.__class__))
        kwargs['obj_type'] = obj_type

        # if no default was specified in the definition, make the default to be
        # an [obj_type] instance
        if 'default' not in kwargs:
            kwargs['default'] = obj_type()

        super().__init__(ObjectType, **kwargs)
        self.description.update(self._get_description(**kwargs))

    def _get_description(self, **kwargs):
        """ Description needs to be json serializable """
        kwargs.update(self._prepare_default(**kwargs))
        kwargs.update(self._prepare_template(**kwargs))
        kwargs['obj_type'] = kwargs['obj_type'].__name__
        return kwargs

    def _prepare_template(self, **kwargs):
        # add object description
        return {"template": self.kwargs["obj_type"]().get_description()}

    def _prepare_default(self, **kwargs):
        """ default in description should be serializable """
        default = kwargs.get('default', None)
        if isinstance(default, PropertyHolder):
            default = default.to_dict()
        return {"default": default}



from nio.modules.proxy import ModuleProxy


class Request(ModuleProxy):

    """ A class encapsulating HTTP request information.

    Your module implementation will proxy this class with its own
    implementation details. An instance of a request will be passed to
    web handlers.

    """

    def get_body(self):
        """ Retrieves request body

        Returns:
            Request body

        """
        raise NotImplementedError()

    def get_identifier(self):
        """ Retrieves request identifier

        Returns:
            Request identifier

        """
        raise NotImplementedError()

    def get_params(self):
        """ Retrieves request params

        Returns:
            Request params

        """
        raise NotImplementedError()

    def get_header(self, header, default=None):
        """ Retrieves a request header

        Returns:
            Request header

        """
        return self.get_headers().get(header, default)

    def get_headers(self):
        """ Retrieves request headers

        Returns:
            Request headers as a dictionary of the form
            {"[header_name]":"[header_value]"}

        """
        raise NotImplementedError()

    def get_method(self):
        """ Retrieves request method

        Returns:
            Request method

        """
        raise NotImplementedError()



import json
from nio.types.base import Type


class DictType(Type):

    @staticmethod
    def serialize(value, **kwargs):
        """ Convert a value to a JSON serializable value """
        return value

    @staticmethod
    def deserialize(value, **kwargs):
        """ Convert value to dict """
        try:
            # if a string is passed, attempt to convert it to a dict
            if isinstance(value, str):
                return json.loads(value)
            else:
                return dict(value)
        except:
            raise TypeError("Unable to cast value to a dict: {}".format(value))



"""
    NIO Command Holder class

"""
from nio.command.base import Command, InvalidCommandArg


class CommandHolder(object):

    @classmethod
    def get_commands_entry(cls):
        """Get all commands on this CommandHolder.

        This does not include commands defined on parent classes.

        Returns:
            dict: commands defined in the class

        """
        commands_attr = "{0}_commands_entry".format(cls.__name__)
        if not hasattr(cls, commands_attr):
            setattr(cls, commands_attr, {})
        return getattr(cls, commands_attr)

    @classmethod
    def get_commands(cls):
        """Get all commands on this CommandHolder and its parent classes.

        This does include commands defined on parent classes.

        Once this method is called, it assumes no new command definitions
        are added to the class.

        Returns:
            dict: all available commands for this class

        """
        commands_attr = "{0}_commands".format(cls.__name__)
        if not hasattr(cls, commands_attr):
            commands = dict()
            commands.update(cls.get_commands_entry())
            for _class in cls.__bases__:
                if issubclass(_class, CommandHolder):
                    commands.update(_class.get_commands())

            setattr(cls, commands_attr, commands)
        return getattr(cls, commands_attr)

    @classmethod
    def add_command(cls, command):
        """Add a command to this CommandHolder.

        Commands are defined as instance methods on the block but
        are added to the block class as 'Command' objects for bookkeeping
        convenience and data encapsulation.

        Args:
            command (Command): A command object to be added.

        """
        class_commands = cls.get_commands_entry()
        if isinstance(command, Command):
            class_commands[command.id] = command
        else:
            raise TypeError("Added command must be an instance of Command")

    @classmethod
    def get_command_description(cls):
        """Get the command descriptions for this CommandHolder and its parents.

        This is useful in serialization/deserialization.

        Returns:
            description (dict): a dictionary containing both property and
                command data.

        """
        class_commands = cls.get_commands()
        commands = {}
        for c in class_commands:
            commands[c] = class_commands[c].get_description()
        return commands

    def invoke(self, id, args):
        """ Call the instance method 'id' with the specified arguments.

        Args:
            id (str): the id of the command/instance method
            args (dict{name: value}): the arguments to the command method.
                These will be assigned to the formal parameters of the
                command, being converted to the appropriate type in the process

        Returns:
            result: the return value of the command method

        """
        class_commands = self.get_commands()
        command = class_commands.get(id)
        if command is None or command.method is None:
            raise RuntimeError("Invalid command: {}".format(id))

        method = getattr(self, command.method, None)

        if method is not None:
            args, kwargs = command.collect_arguments(args)
            try:
                result = method(*args, **kwargs)
            except TypeError as e:
                raise InvalidCommandArg(e)
        else:
            raise RuntimeError("Invalid command: {}".format(id))

        return result



from nio.modules.module import Module
from nio.modules.web import WebEngine, WebServer
from nio.modules.web.http import Request, Response


class WebModule(Module):

    def proxy_web_engine_class(self, web_engine_cls):
        WebEngine.proxy(web_engine_cls)

    def proxy_web_server_class(self, web_server_cls):
        WebServer.proxy(web_server_cls)

    def proxy_request_class(self, request_cls):
        Request.proxy(request_cls)

    def proxy_response_class(self, response_cls):
        Response.proxy(response_cls)

    def finalize(self):
        WebEngine.unproxy()
        WebServer.unproxy()
        Request.unproxy()
        Response.unproxy()
        super().finalize()

    def get_module_order(self):
        return 40

    def get_module_type(self):
        return "web"



from nio.modules.proxy import ModuleProxy


class Subscriber(ModuleProxy):

    """ An interface for a generic Subscriber in the communication module """

    def __init__(self, handler, topic=None, **kwargs):
        """ Create a new subscriber instance.

        Args:
            handler: Specifies where signals will be received. A signature
                definition sample is provided below.

                def process_signals(self, signals)

            topic (str): Defines the kind of information to subscribe to.

            kwargs: Key value argument pairs, used for backwards compatibility.
        """
        super().__init__(handler, topic=topic, **kwargs)

    def open(self, on_connected=None, on_disconnected=None):
        """ Open a subscriber

        Args:
            on_connected (callable): function receiving notification when
                connection is established
            on_disconnected (callable): function receiving notification when
                a disconnection occurs

        When opening a subscriber the system will look for matches against
        existing publishers, when a match is found, a match is made
        so that signals published by matching publisher are received by this
        subscriber

        """
        raise NotImplementedError()

    def is_connected(self):
        """ Determine if this subscriber is connected and ready

        """
        raise NotImplementedError()

    def close(self):
        """ Close a subscriber.

        After closing it is expected all connections to publishers to be
        severed.

        """
        raise NotImplementedError()



from collections import defaultdict
from threading import Lock
from nio.properties.timedelta import TimeDeltaProperty
from nio.modules.scheduler import Job


class Collector(object):

    """ A block mixin that provides collection/buffering functionality.

    By including this mixin, your block will have a `collect` property
    added to it which represents the amount of time to collect for. Any
    subsequent calls to notify signals will not be notified to the block
    router until the collection window is finished.

    This is useful for high volume blocks that wish to buffer their output.

    By setting the collect property to 0, notifying signals will happen
    immediately, as if this mixin wasn't even included.
    """

    collect = TimeDeltaProperty(
        title='Collect Timeout', default={"seconds": 1}, advanced=True, order=100
    )

    def __init__(self):
        super().__init__()
        self._collect_job = None
        self._collect_lock = Lock()
        self._sigs_out = defaultdict(list)

    def start(self):
        # Start the collection job, if we want to be collecting
        if self._are_we_collecting():
            self._collect_job = Job(self._dump_signals, self.collect(), True)
        super().start()

    def stop(self):
        if self._collect_job:
            self._collect_job.cancel()
        super().stop()

    def notify_signals(self, signals, output_id=None):
        """Override the notify signals call to keep collecting"""
        if self._are_we_collecting():
            with self._collect_lock:
                self._sigs_out[output_id].extend(signals)
        else:
            super().notify_signals(signals, output_id)

    def _are_we_collecting(self):
        """Return True if we should be collecting signals"""
        return self.collect().total_seconds() > 0

    def _dump_signals(self):
        """Notify any signals we have collected this window.

        This gets called by the scheduled Job.
        """
        with self._collect_lock:
            for output_id, output_sigs in self._sigs_out.items():
                if len(output_sigs):
                    # Notify the signals for this output ID then
                    # remove the signals from the list
                    super().notify_signals(output_sigs, output_id)
                    del output_sigs[:]



from nio.types.base import Type


class BoolType(Type):

    @staticmethod
    def serialize(value, **kwargs):
        """ Convert a value to a JSON serializable value """
        return value

    @staticmethod
    def deserialize(value, **kwargs):
        """ Convert value to bool"""
        try:
            if isinstance(value, str) and \
                    value.lower() in ['true', 'yes', 'on']:
                return True
            elif isinstance(value, str) and \
                    value.lower() in ['false', 'no', 'off']:
                return False
            return bool(value)
        except:
            raise TypeError("Unable to cast value to bool: {}".format(value))



from nio.testing.modules.web.server import TestingWebServer


class TestingWebEngine(object):

    @classmethod
    def add_server(cls, port, host, config=None):
        return TestingWebServer()

    @classmethod
    def remove_server(cls, server):
        pass

    @classmethod
    def start(cls):
        """ Starts Web Server using configured params """
        pass

    @classmethod
    def stop(cls):
        """ Stops running Web Servers """
        pass

    @classmethod
    def block(cls):
        """ Blocks the Web Server until further request arrives """
        pass



from nio.properties.exceptions import AllowNoneViolation
from nio.properties.util.evaluator import Evaluator


class PropertyValue:
    """ Returned when accessing properties on property holders

    PropertyValues are callable and return the deserialized value of the
    property. If the value is a string that is a valied n.io expression, it
    is first evaluated, optionally against a Signal.

    """

    def __init__(self, property, value=None):
        self._property = property
        self.value = value
        self.evaluator = Evaluator(str(value))

    def __call__(self, signal=None):
        """ Return value, evaluated if it is an expression """
        from nio.properties import PropertyHolder
        if self._property.is_expression(self.value):
            # Expression properties need to be evaluated
            value = self.evaluator.evaluate(signal)
            if value is None:
                if self._property.allow_none:
                    return None
                else:
                    raise AllowNoneViolation("Property value expression is not"
                                             "allowed to evaluate to None")
            else:
                # Deserialize should not be called with None
                return self._property.deserialize(value)
        elif self.value is not None and isinstance(self.value, PropertyHolder):
            # Return property holders as they are
            return self.value
        elif self.value is not None:
            # Deserialize properties
            return self._property.deserialize(self.value)
        elif self.value is None and self._property.allow_none:
            # Return None if it is allowed
            return None
        else:
            raise AllowNoneViolation("Property value None is not allowed")



from nio.modules.context import ModuleContext
from nio.testing.modules.scheduler.job import JumpAheadJob
from nio.testing.modules.scheduler.scheduler import JumpAheadScheduler
from nio.modules.scheduler.module import SchedulerModule


class TestingSchedulerModule(SchedulerModule):

    def initialize(self, context):
        super().initialize(context)
        # For testing, use a job class that allows us to jump ahead in time
        self.proxy_job_class(JumpAheadJob)

        JumpAheadScheduler.do_configure(context)
        JumpAheadScheduler.do_start()

    def finalize(self):
        JumpAheadScheduler.do_stop()
        super().finalize()

    def prepare_core_context(self):
        context = ModuleContext()
        # set a fine resolution during tests
        context.min_interval = 0.01
        context.resolution = 0.01
        return context



from nio.modules.web.handlers.web import WebHandler


class StaticHandler(WebHandler):

    def __init__(self, route):
        """ Create a new StaticHandler instance.

        Args:
            route: uri path

        """
        super().__init__(route)

    def handle(self, request, response):
        """ Handles given request

        Args:
            request: web request
            response: web response

        Returns:
            HTTP response
        """
        raise NotImplementedError()



class RouterContext(object):

    """ Maintains information needed to configure a block router """

    def __init__(self, execution, blocks, settings=None,
                 mgmt_signal_handler=None,
                 instance_id=None,
                 service_id=None,
                 service_name=""):
        """Initializes a router context.

        Args:
            execution (list):  execution list as defined by BaseService
                execution property
            blocks (dict):  dictionary of blocks that looks like this:
                [block_id]: [block instance]
            settings (dict): router settings, these can include
                "clone_signals" and/or any other settings depending on router
                being used
            mgmt_signal_handler (method): method to use to notify
                management signals, receives signal as only parameter
            instance_id: Instance the service belongs to
            service_id (str): service this router belongs to
            service_name (str): The name of the service this router belongs to
        """
        self.execution = execution
        self.blocks = blocks
        self.settings = settings or {}
        self.mgmt_signal_handler = mgmt_signal_handler
        self.instance_id = instance_id
        self.service_id = service_id
        self.service_name = service_name



from collections import defaultdict


class Settings(object):

    """ Helper class to store settings """

    _settings = defaultdict(dict)

    @classmethod
    def get(cls, section=None, option=None, *args, **kwargs):
        """Accessor for settings.

        Args:
            section (str): section to get information from
            option (str): option to get information from

        if no section is specified, all sections are returned
        if no option is specified, all options within section are returned

        Keyword Args:
            fallback: default to use when no information can be found

        Returns:
            settings (AttributeDict): The requested settings or
                None if no possible value was found
        """
        if section is None:
            return cls._settings
        elif option is None:
            return cls._settings.get(section, kwargs.get("fallback", None))
        else:
            try:
                return cls._settings.get(section)[option]
            # TypeError is raised when section is not present
            # KeyError is raised when option is not present
            except (TypeError, KeyError):
                return kwargs.get("fallback", None)

    @classmethod
    def getint(cls, section, option, *args, **kwargs):
        # a convenience method converting section value to an integer
        value = cls.get(section, option, *args, **kwargs)
        if value is not None:
            value = int(value)
        return value

    @classmethod
    def getfloat(cls, section, option, *args, **kwargs):
        # a convenience method converting section value to a float
        value = cls.get(section, option, *args, **kwargs)
        if value is not None:
            value = float(value)
        return value

    @classmethod
    def getboolean(cls, section, option, *args, **kwargs):
        # a convenience method converting section value to a boolean
        value = cls.get(section, option, *args, **kwargs)
        if value is not None:
            value = bool(value)
        return value

    @classmethod
    def getdict(cls, section, option, *args, **kwargs):
        return cls.get(section, option, *args, **kwargs)

    @classmethod
    def set(cls, section, option=None, value=None):
        """ Sets the value for a given section and option

        if the section does not exists, it creates it
        if option is not specified, value is assigned to the section and it
            must be a dictionary type

        Args:
            section (str): section to set information to
            option (str): option to set information to
            value: value to assign

        Raises:
            TypeError if value is invalid
        """
        if option is None:
            cls._settings[section] = value
        else:
            cls._settings[section][option] = value

    @classmethod
    def clear(cls):
        """ Reset settings """
        cls._settings.clear()



import re
from weakref import WeakKeyDictionary

from nio.properties.exceptions import AllowNoneViolation
from nio.properties.util.property_value import PropertyValue


ENVIRONMENT_VAR = re.compile("\[\[([^\[\]]*)\]\]")


class BaseProperty(object):
    """ Base class for metadata property definitions

    Used as a class attribute on property holders.

    Attributes:
        type (Type): The type of data held in this property.
        description (dict): Property settings.
    """

    def __init__(self, _type, title=None, advanced=False, visible=True, 
                    order=None, allow_none=False, default=None, **kwargs):
        self.type = _type

        # make sure title is valid
        if isinstance(title, str) and len(title):
            self.title = title
        else:
            raise ValueError("Title must be a non-empty string")

        if isinstance(order, int) or order is None:
            self.order = order
        else:
            raise ValueError("Order must be an integer")

        self.advanced = advanced
        self.visible = visible
        self.allow_none = allow_none
        self.kwargs = kwargs

        # Default value info
        self._default = default
        self._cached_default = None
        self._default_property_value = PropertyValue(self, self._default)

        # Set up a values dictionary to keep track of values per instance
        # We want a WeakKeyDict so that instances aren't kept in memory
        # based on this property
        self._values = WeakKeyDictionary()

        # Description needs to be serializble so save type as __name__
        self.description = dict(type=_type.__name__,
                                title=title,
                                advanced=advanced,
                                order=order,
                                visible=visible,
                                allow_none=allow_none,
                                default=default,
                                **kwargs)

    @property
    def default(self):
        """ default deserialized value, not a callable PropertyValue """
        if self._cached_default is None and self._default is not None:
            self._cached_default = \
                self.deserialize(self._default, **self.kwargs)
        return self._cached_default

    def __get__(self, instance, owner):
        """ Return the PropertyValue

        Overriding __set__ and __get__ so values can be saved and retrieved
        as callable PropertyValues instead of the raw value.

        Returns:
            PropertyValue: The value for the given instance

            If a value has not been set, return a PropertyValue with the
            default value. If a value has not been set and a default is not
            defined, return None.

        Raises:
            Exception: If a value has not been set and a default is not defined
            and allow_none is False.

        """
        return self._values.get(instance, self._default_property_value)

    def __set__(self, instance, value):
        """ Save the value as a PropertyValue

        Overriding __set__ and __get__ so values can be saved and retrieved
        as callable PropertyValues instead of the raw value.

        """
        self._values[instance] = PropertyValue(self, value)

    def __str__(self):
        return "type is: {}, args are {}".format(self.type, self.kwargs)

    def serialize(self, instance, **kwargs):
        """ Serialze and return the value of an instance of this property

        Serialize is defined by the Type associated with this Property.

        Args:
            instance (PropertyHolder): PropertyHolder that the property is on
        Keyword Args:
            Optionally override the kwargs of the property
            TODO: give an example of when this is useful
        Returns:
            Serialized version of property value

        """
        # Allow property kwargs to be overriden by call to serialize
        merged_kwargs = self.kwargs.copy()
        merged_kwargs.update(**kwargs)
        property_value = self.__get__(instance, instance.__class__)
        # Get the raw value from the PropertyValue
        value = property_value.value
        if value is not None and \
                not self.is_expression(value) and not self.is_env_var(value):
            return self.type.serialize(value, **merged_kwargs)
        else:
            # If the value is None or if it's an expression property,
            # then don't try to serialize it
            return value

    def deserialize(self, value, **kwargs):
        """ Deserialize a value of this property

        Deserialize is defined by the Type associated with this Property.

        Args:
            value: value to be deserialized. Various data formats are accepted
                based on the property Type.
        Keyword Args:
            Optionally override the kwargs of the property
            TODO: give an example of when this is useful
        Returns:
            Deserialized version of property value
        Raises:
            TypeError: value is invalid for Type and cannot be deserialized
            AllowNoneViolation: value is None and allow_none is False

        """
        if not self.is_expression(value) and not self.is_env_var(value):
            # Allow property kwargs to be overriden by call to serialize
            merged_kwargs = self.kwargs.copy()
            merged_kwargs.update(**kwargs)
            if value is None and not self.allow_none:
                raise AllowNoneViolation("Property value None is not allowed")
            return self.type.deserialize(value, **merged_kwargs)
        return value

    def is_expression(self, value):
        """ A property value is an expression if contatins expression syntax

        An expression is identified by a pair of opening curly braces and then
        later a pair of closing curly braces.

        Returns:
            bool: True if value is an expression

        Examples:
            >>> is_expression("{{ 1 + 2 }}")
            True
            >>> is_expression("1 + 2")
            False
            >>> is_expression("{{ 1 + 2")
            False
            >>> is_expression("{ { 1 + 2 } }")
            False

        """

        try:
            # TODO: This check should actually check if the closing braces come
            # after the opening ones.
            return "{{" in value and "}}" in value
        except:
            return False

    def is_env_var(self, value):
        """ Determine if property value is an environment variable

        Environment variables are identified by a pair of opening square braces
        that are eventually followed by a pair of closing square braces.

        It is significant to differently identify a value *containing* an env
        var and a value that *is* an env var. If a value *is* an env var then
        it is valid for that value to be string even if the proprety type can
        not normally be represented as a string.

        Returns:
            bool: True if value is an environment variable

        Examples:
            >>> is_env_var("[[ENV_VAR]]")
            True
            >>> is_env_var("Not an [[ENV_VAR]]")
            False

        """

        try:
            return ENVIRONMENT_VAR.fullmatch(value) is not None
        except:
            return False



from nio.modules.proxy import ModuleProxy


class Persistence(ModuleProxy):

    """ Persistence Module

    This class encapsulates the user-facing interface to NIO's
    persistence layer.

    Block writers use this module to save dynamic data at runtime, allowing
    it to persist in the block instance after a service/instance restart.

    Example:
        self.persistence.save(item, id) # saves the item
        item = self.persistence.load(id) # item is retrieved

    """

    def __init__(self):
        """ Create a Persistence object
        """
        super().__init__()

    def load(self, id, collection=None, default=None):
        """ Load an item from the persistence store.

        Args:
            id (str): Specifies the identifier of the item to load.
            collection (str): if provided, it specifies the collection the item
                belongs to.
            default: the value to return if the item does not exist

        Returns:
            item: The item associated with given id
        """
        raise NotImplementedError()

    def load_collection(self, collection, default=None):
        """ Load a collection from the persistence store.

        Args:
            collection (str): Specifies the collection to load
            default: the value to return if the collection does not exist

        Returns:
            items: The items associated with given collection
        """
        raise NotImplementedError()

    def save(self, item, id, collection=None):
        """ Save the item to the persistence store.

        Args:
            item: Item to save
            id (str): Specifies the identifier of the item to save.
            collection (str): if provided, it specifies the collection the item
                belongs to.
        """
        raise NotImplementedError()

    def save_collection(self, items, collection):
        """ Save a collection to the persistence store.

        Args:
            items: Items to save
            collection (str): Specifies the collection to save
        """
        raise NotImplementedError()

    def remove(self, id, collection=None):
        """ Remove an item from the persistence store.

        Args:
            id (str): Specifies the identifier of the item to remove.
            collection (str): if provided, it specifies the collection the item
                belongs to.
        """
        raise NotImplementedError()

    def remove_collection(self, collection):
        """ Remove a collection from the persistence store.

        Args:
            collection (str): Specifies the collection to remove
        """
        raise NotImplementedError()



from nio.command.params.base import Parameter
from nio.types import DictType


class DictParameter(Parameter):

    def __init__(self, name, **kwargs):
        super().__init__(DictType, name, **kwargs)



from nio.util.versioning.check import is_version_valid, InvalidVersionFormat
from nio.properties import StringProperty


class VersionProperty(StringProperty):

    """ Defines a Version property.

    Version property permits saving version information with
    (major.minor.build) format
    """

    def __init__(self, version=None, title="Version", advanced=True, **kwargs):
        """ Initializes a version property.

        Keyword Args:
            Property definitions

        """
        if "default" in kwargs:
            if version is not None:
                # hmm, both version and default specified,
                # version argument overrides
                kwargs["default"] = version
            super().__init__(title=title, advanced=advanced, **kwargs)
        else:
            super().__init__(title=title, advanced=advanced, default=version, **kwargs)

    def __set__(self, instance, value):
        """ Override default set to make sure it's a valid version """
        # make sure value follows "major,minor,build" convention
        if not is_version_valid(value):
            raise InvalidVersionFormat("Version: {0} is invalid".format(value))

        super().__set__(instance, value)



from nio.properties import BaseProperty
from nio.types import TimeDeltaType


class TimeDeltaProperty(BaseProperty):

    def __init__(self, **kwargs):
        super().__init__(TimeDeltaType, **kwargs)



""" The data that a block will be configured with """


class BlockContext(object):

    def __init__(self, block_router, properties,
                 service_id=None, service_name='', command_url='',
                 mgmt_signal_handler=None):
        """ Initializes information needed for a Block

        This BlockContext will be passed to the `configure` method on each
        block instance before getting started.

        Args:
            block_router (BlockRouter): The router in which the
                block will be running. The only requirements are that it
                can handle signals notified by its blocks.

            properties (dict): The block properties (metadata) that
                will be deserialized and loaded as block properties.
            service_id (str): The service this block belongs to
            service_name (str): The name of the service this block belongs to
            command_url (str): The URL at which this block can be commanded.
                This URL will not have host or port information, as that may
                be different based on public/private IP. It will look like
                "/services/ServiceName/BlockAlias/"
            mgmt_signal_handler (callable): function to use to notify
                management signals, receives signal as only parameter

        """

        self.block_router = block_router
        self.properties = properties
        self.service_id = service_id
        self.service_name = service_name
        self.command_url = command_url
        self.mgmt_signal_handler = mgmt_signal_handler



"""

  A base class and exceptions for proxies.

"""
from collections import defaultdict
from inspect import getmembers, isclass, isfunction, ismethod, isroutine
from nio.util.logging import get_nio_logger


class ProxyNotProxied(Exception):

    """ Exception raised when an operation takes place on an unproxied proxy.

    This can occur when trying to unproxy a proxy that hasn't been proxied yet.
    """
    pass


class ProxyAlreadyProxied(Exception):

    """ Exception raised when an operation takes place on a proxied proxy.

    This can occur when trying to proxy a proxy that has already been proxied
    """
    pass


class ModuleProxy(object):

    """ A base class for creating a ModuleProxy interface

  A ModuleProxy is similar to an interface - it allows for separating the
  interface for accessing a class and the implementation controlling how it
  works.

  To create a ModuleProxy interface, create a class that inherits from the
  ModuleProxy class. Define methods and class variables in your interface with
  the method signatures you want people to use when calling them. These
  functions can have as much or as little functionality as you want. Once the
  proxy is "proxied" with an implementation class, any methods defined on the
  implementation class will be proxied and overridden on to the interface.

  After the interface is proxied, people can create an object as if they are
  creating the interface - the caller does not need to know the type or
  location of the implementation class.

  To create an implementation for a ModuleProxy, create a class that does NOT
  inherit from ModuleProxy. Define functionality for whatever methods on the
  interface you want. You can define additional methods in your class that can
  be accessed by the other methods in your implementation. Be aware though that
  since these methods are not on the interface, it should not be assumed that
  the caller can call them or even knows they exist.

  Once the implementation class is complete, call the proxy method on the
  interface passing in the reference to the implementation class.

  Example - this will proxy the members of the ImplementationClass on to the
  InterfaceProxyClass:

      >>> InterfaceProxyClass.proxy(ImplementationClass)

    """

    # Whether or not this class has already been proxied
    proxied = False
    _impl_class = None
    _unproxied_methods = defaultdict(dict)

    def __init__(self, *args, **kwargs):
        """ Handling the instantiation of a ModuleProxy

        Instantiating a ModuleProxy probably means they want to instantiate
        the module implementation class instead. We still call the ModuleProxy
        constructor so that the interface can define an explicit signature
        for its constructor.

        Therefore, a proxy interface should define its __init__ method with
        the desired signature, and call super with the same arguments. The
        __init__ method of the proxy implementation will NOT be proxied to the
        interface.
        """
        if self.proxied and isclass(self._impl_class):
            self._impl_class.__init__(self, *args, **kwargs)
        else:
            # do not allow creation of not-proxied instances, since allowing
            # such creation yields to unexpected behaviour when after proxying
            # class a proxied method on the instance is invoked.
            raise ProxyNotProxied(
                "An instance of '{0}' class cannot be created, class has not "
                "been proxied".format(self.__class__.__name__))

    @classmethod
    def proxy(cls, class_to_proxy):
        """ Initialize a ModuleProxy by proxying any methods

        This will proxy the class by applying methods from the class passed
        in to the method to the proxy interface. This method should be called
        before any proxied methods are called.

        Args:
            class_to_proxy (class): A reference to the class to proxy on top
                of this interface

        Raises:
            ProxyAlreadyProxied: If the proxy has already been proxied
        """
        if cls.proxied:
            raise ProxyAlreadyProxied()

        # Iterate through the members of the proxy implementation class
        for (name, impl_member) in getmembers(class_to_proxy):
            # Make sure this is a method we want to proxy
            if not cls._is_proxyable(name, impl_member):
                continue

            interface_member = getattr(cls, name, None)
            get_nio_logger("ModuleProxy").debug(
                "Proxying member {0} from {1}".format(
                    name, class_to_proxy.__name__))
            # Save a reference to the original member to replace during unproxy
            cls._unproxied_methods[cls.__name__][name] = interface_member
            setattr(cls, name, impl_member)

        # Mark the class as proxied and save the implementation class
        cls.proxied = True
        cls._impl_class = class_to_proxy

    @classmethod
    def unproxy(cls):
        """ Return the ModuleProxy to its original class methods

        Raises:
            ProxyNotProxied: If the proxy has not yet been proxied
        """
        if not cls.proxied:
            raise ProxyNotProxied()

        for name, iface_member in cls._unproxied_methods[cls.__name__].items():
            if iface_member is None:
                # We didn't have this member on the original interface, delete
                delattr(cls, name)
            else:
                # We had this member originally, replace it with that one
                setattr(cls, name, iface_member)

        # Reset all of our cached proxy class information
        cls._unproxied_methods[cls.__name__] = {}
        cls._impl_class = None
        cls.proxied = False

    @classmethod
    def _is_proxyable(cls, name, member):
        """Returns True if a member is proxy-able.

        Here is what we want to proxy:
            * Any non-private instance functions
            * Any non-private class methods
            * Any non-private class variables
        """
        return not name.startswith('__') and \
            (isfunction(member) or ismethod(member) or not isroutine(member))



from enum import Enum
from threading import Event
from nio.properties import PropertyHolder, ObjectProperty, BoolProperty, \
    IntProperty, SelectProperty, FloatProperty
from nio.block.mixins.retry.strategy import BackoffStrategy
from nio.block.mixins.retry.strategies import LinearBackoff, ExponentialBackoff


class RetryStrategies(Enum):
    linear = LinearBackoff
    exponential = ExponentialBackoff


class RetryOptions(PropertyHolder):
    """ Options the block can be configured with to control how it retries.

    The properties will be passed to the backoff strategy's constructor.
    """

    strategy = SelectProperty(RetryStrategies, title="Strategy to Use",
                              default=RetryStrategies.linear, allow_expr=False)
    max_retry = IntProperty(title="Max Number of Retries", default=5,
                            allow_expr=False)
    multiplier = FloatProperty(title="Retry Multiplier", default=1,
                               allow_expr=False)
    indefinite = BoolProperty(title="Continue Indefinitely?", default=False,
                              allow_expr=False)

    def get_options_dict(self):
        return {
            "max_retry": self.max_retry(),
            "multiplier": self.multiplier(),
            "indefinite": self.indefinite()
        }


class Retry(object):
    """ A block mixin that provides retry functionality.

    By including this mixin, your block will have access to a method that can
    retry on failure. This is useful when performing tasks that have a chance
    of failing but can then work upon retrying. Example use cases are database
    queries or other requests over a network.

    When this mixin is added to a block, some hidden retry configuration
    options will be added to the block. These options allow the block
    configurer to determine how retries should occur, including what strategy
    should be used.

    Block developers can implement their own backoff strategies and employ
    those instead by overriding the setup_backoff_strategy method.

    How to use this mixin:
        1. Configure your block by selecting a backoff strategy as well as
        providing some options to determine how long it will wait between
        retries.

        2. Call execute_with_retry with the function that you want to execute.
        If the target function raises an exception, it will retry until it
        either succeeds or the backoff strategy has decided it should stop
        retrying. If that occurs, execute_with_retry will raise the exception
        that the target function raised originally.

        3. Optionally, override before_retry to write custom code that will be
        performed before attempting a retry.

    Block parameters:
        strategy (select): A choice of pre-configured backoff strategies
        max_retry (int): The maximum number of retries to attempt. Note that
            this is based on retry number, not retry duration. Setting it to
            0 means that no retries will be attempted. Setting it to a negative
            number means that there is no maximum. Also note that this
            property works in concert with the indefinite flag. If that flag
            is set to True, then the retry duration for the retry number
            specified in max_retry will be retried indefinitely.
        multiplier (float): This property has slightly different meanings
            based on what strategy is being used, but for the most part, it
            allows you to control how much time will elapse between retries.
            The higher the number, the longer amount of time that will elapse
            between each retry attempt.
        indefinite (bool): Set to True if you wish for the max_retry retry
            attempt to be continued indefinitely. For example, if this is true
            and max_retry is 5, then the 5th retry will continue to be retried
            until the retry is successful or the block is stopped. If this
            flag is set to False, the retry mixin will stop retrying once the
            max_retry retry attempt is reached.
    """

    retry_options = ObjectProperty(RetryOptions, title="Retry Options",
                                   advanced=True, order=100, default=RetryOptions())

    def configure(self, context):
        """ This implementation will use the configured backoff strategy """
        super().configure(context)
        self.setup_backoff_strategy()

    def setup_backoff_strategy(self):
        """ Define which backoff strategy the block should use.

        This implementation will use the selected backoff strategy from the
        configured retry options and pass other configured options as kwargs.

        Block developers can override this function to use a custom backoff
        strategy.
        """
        self.use_backoff_strategy(
            self.retry_options().strategy().value,
            **(self.retry_options().get_options_dict()))

    def execute_with_retry(self, execute_method, *args, stop_retry_event=None,
                           **kwargs):
        """ Execute a method and retry if it raises an exception

        Args:
            execute_method (callable): A function to attempt to execute. The
                function may be called multiple times if retries occur.
            args/kwargs: Optional arguments to pass to the execute method
            stop_retry_event (Event): Event that stops retrying-loop when set

        Returns:
            The result of execute_method upon success.

        Raises:
            Exception: The exception that execute_method raised when the
                backoff strategy decided to stop retrying.
        """
        # verify incoming event type if set
        if stop_retry_event and not isinstance(stop_retry_event, Event):
            raise TypeError("stop_retry_event must be an instance of Event")
        # Save a stringified version of the method's name
        # If it doesn't define __name__, use however we should stringify
        execute_method_name = getattr(
            execute_method, '__name__', str(execute_method))
        backoff_strategy = self.__new_backoff_strategy()
        while not stop_retry_event or not stop_retry_event.is_set():
            try:
                result = execute_method(*args, **kwargs)
                # If we got here, the request succeeded, let the backoff
                # strategy know then return the result
                backoff_strategy.request_succeeded()
                return result
            except Exception as exc:
                self.logger.warning(
                    "Retryable execution on method {} failed".format(
                        execute_method_name), exc_info=True)
                backoff_strategy.request_failed(exc)
                should_retry = backoff_strategy.should_retry()
                if not should_retry:
                    # Backoff strategy has said we're done retrying,
                    # so re-raise the exception
                    self.logger.exception(
                        "Out of retries for method {}.".format(
                            execute_method_name))
                    raise
                else:
                    # Backoff strategy has instructed us to retry again. First
                    # let the strategy do any waiting, then execute any
                    # pre-work before looping and executing the method again
                    backoff_strategy.wait_for_retry()
                    self.before_retry(*args, **kwargs)

    def use_backoff_strategy(self, strategy, *args, **kwargs):
        """ Tell this mixin which backoff strategy to use.

        This method should be called in a block's configure method.

        Args:
            strategy (class): A subclass of BackoffStrategy to use in this
                block when retrying
            args/kwargs: Optional arguments to pass to the constructor of the
                backoff strategy
        """
        if not issubclass(strategy, BackoffStrategy):
            raise TypeError("Backoff strategy must subclass BackoffStrategy")
        self.__backoff_strategy_class = strategy
        self.__backoff_strategy_args = args
        self.__backoff_strategy_kwargs = kwargs

    def __new_backoff_strategy(self):
        strategy = self.__backoff_strategy_class(
            *self.__backoff_strategy_args,
            **self.__backoff_strategy_kwargs)
        strategy.use_logger(self.logger)
        return strategy

    def before_retry(self, *args, **kwargs):
        """ Perform any actions before the next retry occurs.

        Override this in your block to take action before a retry is attempted.
        This would be the function where you would make any reconnections or
        any other behavior that may make the next retry attempt succeed.
        """
        pass  # pragma: no cover



from time import sleep
from nio.block.mixins.retry.strategy import BackoffStrategy


class ExponentialBackoff(BackoffStrategy):

    def wait_for_retry(self):
        """Sleep a certain number of seconds before we do the retry.

        We will wait the multiplier times 2 raised to one less than the
        current retry number.
        So the first retry will be mult*1, then mult*2, then mult*4, etc

        The exception is when our current retry is greater than our max
        number of retries, then we want to use the max number of retries.
        This can happen if the strategy is configured to run indefinitely.
        """
        sleep(self.multiplier * 2**(min(self.retry_num, self.max_retry)-1))



import re

from nio.util.logging import get_nio_logger

# defines allowed characters withing a topic level
LEVEL_ALLOWED_CHARACTERS = "[0-9A-Za-z_\* \-]"
# pre-compiled match for allowed characters withing a level
level_match = re.compile("^{}+$".format(LEVEL_ALLOWED_CHARACTERS))


class InvalidTopic(Exception):
    """ Exception to be raised when a topic is invalid
    """
    pass


def is_topic_type_valid(topic):
    """ Finds out if a given topic type is valid

    Args:
        topic: topic with 'str' expected type

    Returns:
        True if topic type is str, False otherwise

    """
    if not isinstance(topic, str):
        logger = get_nio_logger("Matching")
        if isinstance(topic, dict):
            logger.warning(
                "Specifying topic: {} through a key-value pair is obsolete, "
                "please use new string-like convention".format(topic))
        else:
            logger.warning("Publisher topic: {} is invalid".format(topic))

        return False

    return True


def is_topic_valid(topic):
    """ Finds out if a given topic is valid

    Topic should be validated when instantiating a Publisher or a Subscriber
    matching algorithm assumes topics are valid.

    Args:
        topic: topic to validate

    Returns:
        True if topic is valid, False otherwise
    """
    if not is_topic_type_valid(topic):
        return False

    sub_topics = topic.split('.')
    for sub_topic in sub_topics:
        # validate topic level
        if not level_match.match(sub_topic):
            return False

    return True



"""

    User support

"""


class User(object):

    """ User class

    Represents the logged in User

    """

    def __init__(self, name='Guest', **kwargs):
        super().__init__(**kwargs)
        self.name = name



import logging

from nio.util.nio_time import get_nio_time


class NIOFilter(logging.Filter):

    def filter(self, record):
        """  Filter the record before being logged

        We add niotime and context properties to the record so loggers
        created before the Logging Module initialization can be written
        using nio formatter
        """
        if not hasattr(record, 'niotime'):
            record.niotime = get_nio_time()
            record.context = record.name
        return True



from nio.testing.modules.communication.manager import PubSubManager


class Subscriber(object):

    """ A version of a Subscriber useful for unit-testing """

    def __init__(self, handler, topic=None, **kwargs):
        self.handler = handler
        self.topic = topic

    def open(self, on_connected=None, on_disconnected=None):
        """ Subscribes handler to matching publishers

        Args:
            on_connected (callable): function receiving notification when
                connection is established
            on_disconnected (callable): function receiving notification when
                a disconnection occurs

        """
        PubSubManager.add_subscriber(self)
        # on_connected, on_disconnected is not applicable,
        # testing communication module is always connected

    def is_connected(self):
        """ Determine if this subscriber is connected and ready

        """
        return True

    def close(self):
        """ Closes publisher.

        This method delegates functionality to notify others the
        subscriber removal via a management signal.

        """
        PubSubManager.remove_subscriber(self)



from re import match
from re import compile as re_compile
from sre_constants import error as RegexError

from nio.modules.security.permissions import PermissionsError
from nio.modules.security.permissions.line import PermissionsLine


class Permissions(object):
    """
    Class which can be loaded a JSON block indicating permissions, or set
    on the fly.

    ---- Loading ----
    The incoming JSON block should be formatted as:
    {
        <resource1>: <permissions_string>,
        <resource2>: <permissions_string>,
        <resourceN>: <permissions_string>
    }

    Note that each resource needs to be a valid regular expression, so that
    later when getting a given resource permission through a 'get' operation,
    the incoming resource can be 'matched' against the above definitions using
    regular expression matching.

    The permissions schema works on a white-list theory, no entries for
    specific resources would equate to no access for a user.

    Specific permissions to a resource can be specified by name,
    keeping in mind that a wildcard permission will be "in addition
    to" rather than "overridden by" a specific permission.
    For example:
        "services.MyService1": <permissions_string>,
        "blocks.*": <permissions_string>

    ---- Storing / Loading ----
    Once permissions are loaded, they are stored in this class under the
    <_permissions> member dict with the following structure:
    {
        <resource1>: PermissionsLine(<resource1_value_from_load>),
        <resource2>: PermissionsLine(<resource2_value_from_load>),
        <resourceN>: PermissionsLine(<resourceN_value_from_load>),
    }

    ---- Queries ----
    Querying a permission is done via the get() member function.

    ---- Setting a permission ----
    Setting a permission can additionally be done via the set() method

    """

    def __init__(self, permissions=None):
        """ Initializes a permission instance

        Args:
            permissions (dict): Initial permissions to load
        """

        # Default to no permissions.
        self._permissions = {}

        # Load permissions if specified
        if permissions:
            self.load(permissions)

    def load(self, permissions):
        """ Loads a dictionary of permissions

        Args:
            permissions (dict): permission specifications

        Raises:
            PermissionError if an invalid permission spec. is found
        """
        # Check that the permissions block has the proper structure.
        if not isinstance(permissions, dict):
            raise PermissionsError("No valid 'permissions' specified.")

        # For each key in the incoming block, parse the permissions out into
        # the proper structure.
        for regex, perm_str in permissions.items():

            try:
                # validate regex at this level
                re_compile(regex)
                self._permissions[regex] = PermissionsLine(perm_str)
            except RegexError:
                raise PermissionsError("'{0}' is not a valid regex for "
                                       " matching permissions.".format(regex))

    def get(self, resource, permission):
        """ Finds out if resource has given permission

        Args:
            resource: Specific service to request permission for.
            permission: The permission to check.

        Returns:
            True if permission is present, otherwise False.
        """

        # Ensure the permission being requested is valid, mapping to the
        # shorthand permission if necessary.
        permission = permission.lower()
        if permission in PermissionsLine.SHORTHAND:
            permission = PermissionsLine.SHORTHAND[permission]

        for pattern, permissions in self._permissions.items():
            # add a $ to match the end of the string (otherwise it would match
            # sub-strings, i.e., "blocks" would match "blocks.MyBlock")
            m = match("{}$".format(pattern), resource)
            if m and getattr(permissions, permission, False):
                return True
        return False

    def set(self, resource, permission, value=True):
        """ Sets a permission on given resource

        In addition to loading permissions from a dictionary during
        creation, permissions can be set explicitly

        Args:
            resource (str): Specific resource being assigned a permission.
            permission (str): The permission to check.
            value (bool): Permission value to use

        Raises:
            PermissionError if an invalid permission spec. is specified
        """

        # Make sure the value for the permission is valid (True or False only).
        if value not in [True, False]:
            raise PermissionsError("Unable to assign permission value '{0}'. "
                                   "Must be True or False.".format(value))

        # Ensure the permission being requested is valid, mapping to the
        # shorthand permission if necessary.
        permission = permission.lower()
        if permission in PermissionsLine.SHORTHAND:
            permission = PermissionsLine.SHORTHAND[permission]

        try:
            # make sure resource is a valid regex
            re_compile(resource)
        except RegexError:
            raise PermissionsError("'{0}' is not a valid regex for "
                                   " matching permissions.".format(resource))

        if resource not in self._permissions:
            self._permissions[resource] = PermissionsLine()

        # assign actual permission
        setattr(self._permissions[resource], permission, value)

    def to_dict(self):
        """ Return a serializable dictionary of the permissions

        Returns:
            dict: keys are resource permission regexes, values are shorthand
            strings representing the permissions on that resource
        """
        return {
            resource_regex: str(line)
            for resource_regex, line in self._permissions.items()
        }

    def update(self, perms):
        """ Update the current Permissions object with another permissions
            object, with the second (argument) taking priority over the
            original if there is a key collision.

            Returns:
               Permissions object.
        """

        self.load(perms.to_dict())
        return self



from nio.project.block import BlockEntity
from nio.project.configuration import ConfigurationEntity
from nio.project.entity import Entity
from nio.project.project import Project
from nio.project.service import ServiceEntity




"""
  NIO block support base class

"""
from uuid import uuid4

from collections import defaultdict
from nio.testing.test_case import NIOTestCase
from nio.router.base import BlockRouter
from nio.block.context import BlockContext
from nio.block.terminals import DEFAULT_TERMINAL


class BlockRouterForTesting(BlockRouter):

    """ A testing router to be used in block test cases.

    Supports block tests by keeping track of signal counts per block and
    block status as provided by management signals.
    """

    def __init__(self):
        super().__init__()
        self._test_case = None
        self._block_signal_counts = {}
        self._block_status = {}

    def configure(self, test_case):
        """ Configures router by saving the test case being supported

        Args:
            test_case (NIOTestCase): test case to call when notifying signals

        """
        self._test_case = test_case

    def notify_signals(self, block, signals, output_id=None):
        """ Receives block signals.

         Keeps track of the signal count per block and per output id
         Forwards notification to test_case's '_internal_signals_notified'

        Args:
            block (Block): notifier block
            signals (list): signals being notified
            output_id: output identifier

        """
        if output_id is None:
            output_id = DEFAULT_TERMINAL

        self._block_signal_counts[block] = \
            self._block_signal_counts.get(block) or defaultdict(int)
        self._block_signal_counts[block][output_id] += len(signals)

        self._test_case._internal_signals_notified(block, signals, output_id)

    def get_signals_from_block(self, block=None, output_id=DEFAULT_TERMINAL):
        """ Provides signal count per block or for all blocks

        Args:
            block (Block): block instance
            output_id: output identifier

        Returns:
            The number of signals a block has notified.
            If block is None, return the total number of signals notified

        """
        if block is None:
            return sum([sum(blk.values()) for blk
                        in self._block_signal_counts.values()])

        block_counts = self._block_signal_counts.get(block)
        if block_counts is None:
            return 0

        return block_counts.get(output_id) or 0


class NIOBlockTestCase(NIOTestCase):

    """ Base Unit Test case class for blocks

    Allows block tests to see the signals that get notified and access some
    other data about them by setting up a router.

    Simply create the block you want to use and call configure_block with the
    properties the block should have.

    Example:
    b = MyTestableBlock()
    self.configure_block(b, {"loglevel": "DEBUG"})

    Now, whenever b notifies signals, the signals_notified function will be
    called in the test case.
    """

    def __init__(self, methodName='runTests'):
        super().__init__(methodName)
        self._router = BlockRouterForTesting()
        self.notified_signals = defaultdict(list)
        # support internal test case control over signal notifications
        self._last_output_notified = None
        # keep track of mgmt signals notified per block
        self._block_mgmt_signal_counts = defaultdict(int)

    def setUp(self):
        super().setUp()
        self._router.configure(self)

    def configure_block(self, block, block_properties):
        """ Configures block by assigning properties and router to given block

        Args:
            block (Block): block instance
            block_properties (dict): properties to assign

        """
        # Blocks should always have an 'id', but we'll let it pass in tests
        block_properties["id"] = block_properties.get("id", uuid4().hex)
        block.configure(BlockContext(
            self._router,
            block_properties,
            'TestSuite',
            '',
            mgmt_signal_handler=self.management_signal_notified))

    @property
    def last_notified(self):
        # keep this property as an alias for backwards compatibility
        # convert to previous form, a list of signals

        return defaultdict(list, ((output_id, [notified_signal for sublist in
                                              self.notified_signals[output_id]
                                              for notified_signal in sublist])
                                  for output_id in self.notified_signals))

    def last_signal_notified(self, output_id=None):
        """ Provides last signal notified

        If no signal has been notified on specified output, the last output
        notified on is used.

        If output_id is specified, it must be valid

        Args:
            output_id (str): output identifier

        Returns:
            last signal notified

        Raises:
            ValueError if output_id is invalid
        """
        # if output_id is specified, it must be valid.
        if output_id is not None and \
           output_id not in self.notified_signals:
            raise ValueError("Invalid output id specified")

        # if a valid output_id is not provided, use last output notified on.
        if output_id is None:
            output_id = self._last_output_notified

        return self.notified_signals[output_id][-1][-1]

    def _internal_signals_notified(self, block, signals, output_id):
        """ Receives internal block signals notification

        Calls overrideable public method 'signals_notified' and updates
        internal members that support block unit testing.

        """
        self._last_output_notified = output_id

        self.signals_notified(block, signals, output_id)

    def signals_notified(self, block, signals, output_id):
        """ Receives block signals notification

        This method can be overriden by block developer

        Args:
            block (Block): notifying block
            signals (list): signals being notified
            output_id: output identifier

        """
        self.notified_signals[output_id].append(signals)

    def management_signal_notified(self, signal):
        """ Receives block management signal notification

        Args:
            signal (Signal): signal being notified
        """
        # update per block signal counting
        self._block_mgmt_signal_counts[signal.block_id] += 1

    def assert_num_signals_notified(self, num, block=None,
                                    output_id=DEFAULT_TERMINAL):
        """ Assert that the number of signals notified is a certain number.

        Args:
            num (int): number to check against
            block (Block): the block which should have notified. If not
                included, all blocks will be considered.
            output_id: The output id of the block to consider.
        """
        signals = self._router.get_signals_from_block(block, output_id)
        self.assertEqual(num, signals)

    def assert_num_mgmt_signals_notified(self, num, block=None):
        """ Assert the number of management signals notified is a number.

        Args:
            num (int): number to check against
            block (Block): the block which should have notified. If not
                included, all blocks will be considered.
        """
        if block:
            self.assertEqual(
                num, self._block_mgmt_signal_counts[block.id()])
        else:
            self.assertEqual(
                num, sum(self._block_mgmt_signal_counts.values()))

    def assert_signal_notified(self, signal, position=None,
                               output_id=DEFAULT_TERMINAL):
        """ Assert that the given signal has been notified from the block 
        
        Args:
            signal (signal object): signal to assert has been notified from 
                the block
            position (int): the position of the signal in the notified_signals 
                list
            output_id (string): output id of the block
        """
        total_notified_signals = [notified_signal.to_dict() for sublist in
                                  self.notified_signals[output_id]
                                  for notified_signal in sublist]
        if position is not None:
            self.assertEqual(signal.to_dict(),
                             total_notified_signals[position])
        else:
            self.assertIn(signal.to_dict(), total_notified_signals)

    def assert_signal_list_notified(self, signal_list, position=None,
                                    output_id=DEFAULT_TERMINAL):
        """ Assert that the given signal list has been notified from the 
        block. The given list must match the order in which the signals were
        notified.
        
        Args:
            signal_list (list): list of signal objects to assert
            output_id (string): output terminal of the block to assert against
            position (int): position of the signal list in 
                notified_signals[output_id]
        """
        input_list = [signal.to_dict() for signal in signal_list]
        notified_lists = [[signal.to_dict() for signal in notified_list]
                          for notified_list in self.notified_signals[output_id]]

        if position is not None:
            self.assertListEqual(input_list, notified_lists[position])
        else:
            self.assertIn(input_list, notified_lists)

    def assert_last_signal_notified(self, signal, output_id=DEFAULT_TERMINAL):
        """ Assert that the last signal notified on the given output_id is
        equal to 'signal'. This is designed to be a convenience function
        for last_signal_notified functionality, and provides dict comparison.
        
        Args:
            signal (signal object): the signal to assert against
            output_id (string): output terminal of the block to assert against
        """
        self.assertDictEqual(self.last_signal_notified(output_id).to_dict(),
                             signal.to_dict())

    def assert_last_signal_list_notified(self, signal_list,
                                         output_id=DEFAULT_TERMINAL):
        """ Assert the given list was notified from the block. Provides dict
        comparison.

        Args:
            signal_list (list): The list of signals to assert was notified
        """
        notified_list = [signal.to_dict() for signal in
                         self.notified_signals[output_id][-1]]
        signal_dict_list = [signal.to_dict() for signal in signal_list]

        self.assertListEqual(notified_list, signal_dict_list)

    def assert_block_status(self, block, status):
        """ Asserts that the given block has a certain block status

        Blocks will have a status of '' if they have not had a status
        reported yet.

        Args:
            block (Block): The block to check the status of
            status (BlockStatus.xxx): The BlockStatus enum value to have
        """
        self.assertEqual(self._router._block_status.get(block, ''), status)



"""

  Authenticator Interface

"""
from nio.modules.proxy import ModuleProxy


class Authenticator(ModuleProxy):

    @classmethod
    def authenticate(cls, request):
        """ Authenticate a request and return a User.

        The objective of this method is to determine the user who is making a
        given web request. Most implementations will probably force the request
        object passed in to be an instance of Request from the web module.

        Args:
            request (Request): A request from the web module used to determine
                the user who is making this request

        Returns:
            User: A reference to the user making the request.

        Raises:
            Unauthorized: If it cannot determine the user making the request
        """
        raise NotImplementedError()



import datetime
import json
import math
import random
import re


class Parser:

    """ Helper class for translation and evaluation of nio expressions

    Args:
        signal (Signal): Contextualizes incoming code snippets.
        default (ANY): If an incoming signal does not have the
            requested attribute, this value is interpolated instead.

    """
    escaped = re.compile(r'\\(\$|{{|}})')
    ident = re.compile(r'(?<!\\)\$([_A-Za-z]([_A-Za-z0-9])*)?|[^\$]*')
    ident_stem = re.compile(r'[_a-zA-Z][_a-zA-Z0-9]*$')

    def parse(self, tokens):
        """ Parse a list of tokens and evaluate them.

        Traverses the 'AST' (list of tokens) recursively, transforming,
        ignoring, or evaluating them as needed.

        Args:
            tokens (list(str)): Tokens corresponding to the expression
                being evaluated.

        """
        expr = ''
        if not tokens:
            result = []
        elif tokens[0] == '{{':
            tokens.pop(0)

            # Gobble up tokens until the closing delimiter
            while tokens[0] != '}}':
                expr += tokens.pop(0)
                if len(tokens) == 0:
                    raise SyntaxError("Unexpected EOF while parsing")
            tokens.pop(0)

            result = self._build_function(expr) + self.parse(tokens)
        else:

            # Just a raw string. Remove any escape characters and
            # move on.
            _next = self.escaped.sub(self._unescape, tokens.pop(0))
            result = [_next] + self.parse(tokens)

        return result

    def _build_function(self, expr):
        """ Build a function that evaluates an expression.

        Helper function to transform incoming expressions and wrap
        them in unnamed functions. This leaves the expressions themselves
        compiled and ready to be parameterized with incoming signals
        """
        transformed = self.ident.sub(self._transform_attr, expr)
        unescaped = self.escaped.sub(self._unescape, transformed)
        try:
            return [eval("lambda signal, self: {}".format(unescaped))]
        except Exception as e:
            _type = type(e)
            raise _type(
                "Error while evaluating {}: {}".format(unescaped, str(e))
            )

    def _unescape(self, match):
        return match.group(0)[1:]

    def _transform_attr(self, match):
        tok = match.group(0).split('$')
        result = tok[0]
        if not result:
            # grab the signal
            result = 'signal'

            # if the rest of the token is a valid identifier, generate code to
            # get the corresponding attribute. otherwise just return the signal
            if self.ident_stem.match(tok[1]) is not None:
                result = 'getattr(signal,"{0}")'.format(tok[1])

        return result



from nio.modules.proxy import ModuleProxy


class Response(ModuleProxy):

    """ A class encapsulating HTTP response information.

    Your module implementation will proxy this class with its own
    implementation details. An instance of a response will be passed to
    web handlers.

    """

    def set_header(self, header_name, header_value):
        """ Sets a single HTTP header.

        Args:
            header_name: header name
            header_value: header value

        """
        raise NotImplementedError()

    def set_status(self, status, message=None):
        """ Sets a web response status.

        Args:
            status: http status
            message: optional message

        """
        raise NotImplementedError()

    def set_body(self, body):
        """ Sets a web response body.

        Args:
            body: response body

        """
        raise NotImplementedError()



from nio.testing.modules.persistence import Persistence
from nio.modules.persistence.module import PersistenceModule


class TestingPersistenceModule(PersistenceModule):

    """ A version of the persistence module used for unit tests.

    This will proxy the Persistence class that does in-memory persistence
    and is usable in most unit tests.
    """

    def initialize(self, context):
        super().initialize(context)
        Persistence.configure(context)
        self.proxy_persistence_class(Persistence)

    def finalize(self):
        super().finalize()



from nio.modules.module import Module
from nio.modules.persistence import Persistence


class PersistenceModule(Module):

    def proxy_persistence_class(self, persistence_class):
        Persistence.proxy(persistence_class)

    def finalize(self):
        Persistence.unproxy()
        super().finalize()

    def get_module_order(self):
        return 10

    def get_module_type(self):
        return "persistence"



from nio.testing.modules.communication.manager import PubSubManager


class Publisher(object):

    """ A version of a Publisher useful for unit-testing """

    def __init__(self, topic=None, **kwargs):
        self.topic = topic

    def open(self, on_connected=None, on_disconnected=None):
        """ Opens publishing channel

        Args:
            on_connected (callable): function receiving notification when
                connection is established
            on_disconnected (callable): function receiving notification when
                a disconnection occurs

        This method delegates functionality to notify others the
        publisher creation and its definitions via a management signal.

        """
        PubSubManager.add_publisher(self)
        # on_connected, on_disconnected is not applicable,
        # testing communication module is always connected

    def is_connected(self):
        """ Determine if this publisher is connected and ready

        """
        return True

    def send(self, signals):
        """ Sends signals

        Args:
            signals: Signals to send
        """
        PubSubManager.send(self, signals)

    def close(self):
        """ Closes publisher.

        This method delegates functionality to notify others the
        publisher removal via a management signal.

        """
        PubSubManager.remove_publisher(self)

    def is_closed(self):
        """ Finds out if publisher has been closed """
        return self not in PubSubManager.publishers



from enum import Enum
from nio.types.base import Type


class SelectType(Type):

    @staticmethod
    def serialize(value, **kwargs):
        """ Convert a value to a JSON serializable value """
        if value is not None and isinstance(value, Enum):
            return value.name
        return str(value)

    @staticmethod
    def deserialize(value, **kwargs):
        """ Convert value to select

        Keyword Args:
            enum (Enum): Enumeration type
        """
        # Value can be an Enum
        if isinstance(value, Enum):
            return value

        # Value can be a string equal the name or value of the Enum
        for name, member in kwargs.get('enum').__members__.items():
            if name == value or member.value == value:
                return member

        raise TypeError("{} does not match enum type: {}".format(
            value, kwargs.get('enum')))



from threading import Lock


class LockQueueFull(Exception):
    pass


class LimitLock(object):

    """ A block mixin that provides lock functionality.

    By including this mixin, your block will have access to methods for lock
    functionality. This is useful when performing tasks that need a lock but
    could potentially take a long time so eventually you want to abort instead
    of having more threads wait for locks.

    How to use this mixin:
        1. Call _execute_with_lock with the method that you want to execute
        and any keyword arguments. The arg max_locks defines how many threads
        will wait for this lock before aborting.

    Note: This method can only be used in one place in your block code.

    """

    def __init__(self):
        super().__init__()
        self._exceute_lock = Lock()
        self._number_of_locks = 0

    def execute_with_lock(self, execute_method, max_locks, *args, **kwargs):
        """ Execute a given method inside of a limited lock

        This method will wait to acquire a lock before executing the specified
        method, but will limit the number of threads who are allowed to wait
        for the lock.

        Args:
            execute_method (function): The function to execute once the lock
                is acquired
            max_locks (int): The maximum number of threads allowed to wait for
                the lock to acquire
            *args, **kwargs: Additional arguments to pass to the execute_method

        Returns:
            The result of the execute_method, once it actually executes

        Raises:
            LockQueueFull: If the call exceeds the maximum number of threads
                that can wait for the lock.
        """
        if self._number_of_locks >= max_locks:
            self.logger.warning(
                "Currently {} locks waiting to be acquired. This is more than "
                "the max. Aborting call to method in this thread.".format(
                    self._number_of_locks))
            raise LockQueueFull
        self._number_of_locks += 1
        with self._exceute_lock:
            try:
                return execute_method(*args, **kwargs)
            finally:
                self._number_of_locks -= 1



""" The definition of a Signal

A signal is the fundamental unit of information in n.io. Functionally, it
is a key-value object that can contain any information that Python can
represent. Conceptually, a signal can represent any piece of information. It
can be as simple as an empty signal representing some sort of ping or as
complex as a signal with many attributes containing various pieces of
information.

The base signal does not impose very many rules or restrictions, but the class
is intended to be extended and sub-classed to have more validation and
functionality for different types of data.
"""
from nio.util.class_attributes import ClassAttributes
from nio.util.logging import get_nio_logger


class Signal(object):

    def __init__(self, attrs=None):
        """ Create a new signal - optionally with some data

        Args:
            attrs (dict): An optional dictionary containing the data for this
                signal. If specified, this will dictionary will be passed to
                from_dict.
        """
        super().__init__()
        if attrs is not None:
            self.from_dict(attrs)

    def from_dict(self, data):
        """ Create attributes on this signal based on a dictionary.

        This function will take a dictionary of data and attempt to set the
        attributes on this signal. This does not always work though because
        some values that are valid keys on a dictionary are not valid
        attribute names on an object (e.g. False, 7, None, etc). In the case
        of an invalid attribute, a warning will be logged and the attribute
        will be skipped.

        Args:
            data (dict): A dictionary of data to set on this signal

        Raises:
            TypeError: If data is not a dictionary
        """
        if not isinstance(data, dict):
            raise TypeError("Signal data from_dict must be a dictionary")

        for key, val in data.items():
            try:
                # We don't want to have the empty string be on the signal as
                # an attribute name, even though Python allows it
                if key and isinstance(key, str):
                    setattr(self, key, val)
                    # If we made it here, keep going to the next attribute
                    continue
            except Exception:
                # In case of an exception, fail gracefully but log a warning
                get_nio_logger('Signal').exception(
                    "Failed to assign value: {} to attribute: {}".
                    format(val, key))

            # If we made it here something was wrong with the key/val,
            # otherwise we would have "continued" through the loop
            message = "Key: {} with value: {} could not be made part of the "\
                      "signal".format(key, val)
            get_nio_logger('Signal').warning(message, exc_info=True)
            raise ValueError(message)

    def to_dict(self, include_hidden=False, with_type=False):
        """ Create a dictionary representation of this signal.

        Args:
            include_hidden (bool): Whether or not to include hidden signal
                attributes in the resulting dictionary. Defaults to False
            with_type (str): If specified, the attribute on the resulting
                dictionary that will be the type of this signal

        Returns:
            dict: A dictionary containing the attributes of the signal
        """
        sig_dict = dict()
        for attr_name in dir(self):

            # We don't want to include attributes starting with two underscores
            # under any circumstance
            if attr_name.startswith('__'):
                continue

            attr_value = getattr(self, attr_name)

            # We don't want to include methods or functions in the dictionary
            if not ClassAttributes.is_attr(attr_value):
                continue

            # We only want the attribute if we want hidden attributes or
            # if it's not hidden
            if include_hidden or not self._is_hidden(attr_name):
                sig_dict[attr_name] = attr_value

        if with_type and isinstance(with_type, str):
            sig_dict[with_type] = self.__class__.__name__

        return sig_dict

    def _is_hidden(self, attribute_name):
        """ Returns True if a given attribute name is hidden.

        Currently, attributes that begin with an underscore are to be
        considered "hidden" attributes. This simply means they will not be
        returned in to_dict calls on this signal unless requested.

        Args:
            attribute_name (str): The name of the attribute to check

        Returns:
            bool: True if the attribute is to be considered hidden
        """
        return attribute_name.startswith('_')

    def __str__(self):
        d = self.to_dict()
        return repr(d) if len(d) else self.__class__.__name__

    def __eq__(self, other):
        if not isinstance(other, Signal):
            return False

        if self.to_dict() == other.to_dict():
            return True
        else:
            return False

    def __hash__(self):
        return hash(frozenset(self.to_dict()))



import logging
from datetime import datetime

from nio.util.logging.handlers.publisher.log_signal import LogSignal
from nio.util.logging.handlers.publisher.proxy import PublisherProxy


class PublisherHandler(logging.Handler):

    """ Publisher handler for logging.

    Publishers log events using the nio Publisher interface therefore
    allowing reception of log events through the network through the
    instantiation of Subscribers

    Instances of this class will be registered with the base Logger on an
    as-needed basis.

    Users of this handler should be aware of the potential risk of issuing
    logging statements within the Subscriber catching signals published through
    this handler, therefore, logging within the subscriber 'handler' is
    not recommended.

    """

    def __init__(self, topic="nio_logging",
                 max_publisher_ready_time=5,
                 publisher_ready_wait_interval_time=0.1):
        """  Create a new PublisherHandler instance.

        Args:
            topic (str): topic to use when publishing log messages
            max_publisher_ready_time (float): maximum time to wait for
                publisher to be ready
            publisher_ready_wait_interval_time (float): interval in seconds to
                use when waiting for publisher to be ready
        """
        super().__init__()

        # Initialize unique proxy for all publisher handlers.
        PublisherProxy.init(topic,
                            max_publisher_ready_time,
                            publisher_ready_wait_interval_time)

    def emit(self, record):
        """ Publish the log record on the opened publisher

        Args:
            record (LogRecord): record to be logged.

        Returns:

        """
        try:
            # publishing it as a signal
            signal = LogSignal(self._get_time_as_str(record.created),
                               record.context,
                               record.levelname,
                               record.msg,
                               record.filename,
                               record.funcName,
                               record.lineno)
            PublisherProxy.publish([signal])
        # allow exceptions like KeyboardInterrupt and SystemExit to propagate
        # and catch NotImplementedError since when stopping a service
        # the 'Publisher' interface is eventually un-proxied
        except NotImplementedError:
            pass

    def close(self):
        """ Closes handler

        Releases/Closes any resources used by handler

        """
        # close all dependencies
        PublisherProxy.close()

        super().close()

    @staticmethod
    def _get_time_as_str(epoch):
        """ Converts incoming epoch time to a str formatted version

        Args:
            epoch (int): number of seconds

        Returns:
            ISO 8601-conformant time string

        """
        d = datetime.utcfromtimestamp(epoch).isoformat() + "Z"
        # convert to str and trim from microseconds to milliseconds ([:-3])
        return d



import inspect
from collections import Iterable
from copy import deepcopy

from nio.router.diagnostic import DiagnosticManager
from nio.signal.base import Signal
from nio.util.runner import Runner, RunnerStatus


class BlockRouterNotStarted(Exception):
    pass


class InvalidBlockInput(Exception):
    pass


class InvalidBlockOutput(Exception):
    pass


class MissingBlock(Exception):
    pass


class InvalidProcessSignalsSignature(Exception):
    pass


class BlockReceiverData(object):

    """ A class that defines block receiver information.

    An instance of this class is created for each block receiver and accessed
    when delivering signals
    """

    def __init__(self, block, input_id, output_id):
        """ Create a new block receiver data.

        Take care of setting up instance variables.
        """
        self.block = block
        self.input_id = input_id
        self.output_id = output_id
        self.include_input_id = self._block_defines_input_id(block)

    def _block_defines_input_id(self, block):
        """ Returns True if the block developer can receive the input ID """
        args_spec = inspect.getargspec(block.process_signals)
        if len(args_spec.args) == 2:
            # method signature assumed to be (self, signals)
            # ==> disregard input_id even when present
            return False
        elif len(args_spec.args) == 3:
            # method signature assumed to be (self, signals, input_id)
            return True
        else:
            raise InvalidProcessSignalsSignature(
                "Block {} signature is invalid".format(block.label(True)))


class BlockRouter(Runner):

    """ A class that can route signals between blocks in a service.

    A Block Router receives service block execution information, processes
    it and becomes ready to receive signals from any participating block in
    the service execution.

    When receiving signals, the block router would be in the position of
    a speedy delivery to receiving blocks.
    """

    def __init__(self):
        """ Create a new block router instance """
        # Parent class Runner creates logger for class
        super().__init__(
            status_change_callback=self._on_status_change_callback)

        self._receivers = None
        self._clone_signals = False
        self._check_signal_type = True
        self._diagnostics = True
        self._diagnostic_manager = None

    def configure(self, context):
        """Configures block router.

        This method receives the block router settings and
        it will based on these automate certain tasks so that
        signal delivery is made in an optimized fashion.

        Args:
            context (RouterContext): Context where settings are stored
        """

        self._clone_signals = \
            context.settings.get("clone_signals", True)
        if self._clone_signals:
            self.logger.info('Set to clone signals for multiple receivers')
        self._check_signal_type = \
            context.settings.get("check_signal_type", True)
        self._diagnostics = \
            context.settings.get("diagnostics", True)
        if self._diagnostics:
            self._diagnostic_manager = DiagnosticManager()
            self._diagnostic_manager.do_configure(context)

        # cache receivers to avoid searches during signal delivery by
        # creating a dictionary of the form
        # {block_id: block instance receivers list}
        self._receivers = {}
        # Go through list of receivers for a given block as
        # defined in "execution" entry, and parses out needed information
        # to be used when delivering signals.
        for block_execution in context.execution:
            # block_execution should be an instance of BlockExecution
            sender_block_id = block_execution.id()
            sender_block = context.blocks[sender_block_id]
            self._receivers[sender_block_id] = []

            # check if receivers have the {output_id: [receivers]} format
            if isinstance(block_execution.receivers(), dict):

                for output_id, block_receivers in \
                        block_execution.receivers().items():

                    # validate that output_id is valid for sender
                    if not sender_block.is_output_valid(output_id):
                        msg = "Invalid output: {} for block: {}".format(
                            output_id, sender_block_id)
                        self.logger.error(msg)
                        raise InvalidBlockOutput(msg)

                    parsed_receivers = \
                        self._process_receivers_list(block_receivers,
                                                     context.blocks,
                                                     output_id)
                    self._receivers[sender_block_id].extend(parsed_receivers)
            else:
                # any receivers specified?
                if block_execution.receivers():
                    # receivers have the 'simple list format', i.e. [receivers]
                    # a default output is mandatory then since no output is
                    # specified when using the 'simple list format'
                    if sender_block._default_output is None:
                        raise InvalidBlockOutput(
                            "Block {} does not define a default output, must "
                            "specify outputs in execution config".format(
                                sender_block))
                    parsed_receivers = self._process_receivers_list(
                        block_execution.receivers(),
                        context.blocks,
                        sender_block._default_output.id)
                    self._receivers[sender_block_id].extend(parsed_receivers)

    def start(self):
        super().start()
        if self._diagnostics:
            self._diagnostic_manager.do_start()

    def stop(self):
        if self._diagnostics:
            self._diagnostic_manager.do_stop()
        super().stop()

    def _process_receivers_list(self, receivers, blocks, output_id):
        """ Goes through and process each receiver

        Args:
            receivers: Receiver definition
            blocks (dict): instantiated service blocks
            output_id: specifies if receiver is for a given output identifier

        Returns:
            parsed list of receivers
        """
        receivers_list = []
        for receiver in receivers:
            receiver_data = self._process_block_receiver(receiver,
                                                         blocks,
                                                         output_id)
            receivers_list.append(receiver_data)

        return receivers_list

    def _process_block_receiver(self, receiver, blocks, output_id):
        """ Process a receiver component from a list of a block's receivers

        Args:
            receiver: Receiver definition
            blocks (dict): instantiated service blocks
            output_id: specifies if receiver is for a given output identifier

        Returns:
            Parsed receiver data
        """

        # determine input_id and block receiver id
        if isinstance(receiver, dict):
            try:
                input_id = receiver["input"]
                receiver_id = receiver["id"]
            except KeyError:
                self.logger.exception("Invalid format while processing block "
                                      "receiver: {}".format(receiver))
                raise
        else:
            if receiver not in blocks:
                raise MissingBlock()
            # handling old format, get a defined input,
            default_input = blocks[receiver]._default_input
            if default_input is None:
                raise InvalidBlockInput("Block does not have a default input")
            input_id = default_input.id
            receiver_id = receiver

        try:
            receiver_block = blocks[receiver_id]
        except KeyError as e:
            self.logger.exception(
                "Missing block for receiver: {}".format(receiver_id))
            raise MissingBlock(e)

        # make sure input is valid for this block
        if not receiver_block.is_input_valid(input_id):
            raise InvalidBlockInput(
                "Input {} is invalid for block {}".format(
                    input_id, receiver_block))

        receiver_data = BlockReceiverData(receiver_block,
                                          input_id,
                                          output_id)
        return receiver_data

    def _on_status_change_callback(self, old_status, new_status):
        self.logger.info("Block Router status changed from: {} to: {}".
                         format(old_status.name, new_status.name))

    def notify_signals(self, block, signals, output_id):
        """This method is called when a block is notifying signals.

        Args:
            block (Block): The block that is notifying
            signals (list): The signals that the block is notifying
            output_id: output identifier

        The signals argument is handled as follows:
            - every signal notified has to be an instance of 'Signal' by
                default, although it is configurable
            - an empty list or something evaluating to False is discarded

        """
        if self.status.is_set(RunnerStatus.started):

            if not signals:
                # discard an empty list or something that evaluates to False
                return

            if output_id is None:
                if block._default_output is None:
                    raise InvalidBlockOutput(
                        "Block does not define a default output, must "
                        "explicitly specify output in notify_signals")
                else:
                    output_id = block._default_output.id
            elif not block.is_output_valid(output_id):
                raise InvalidBlockOutput(
                    "Output {} not defined on block {}".format(
                        output_id, block))

            # make sure we can iterate
            if not isinstance(signals, Iterable):
                raise TypeError("Block Router must be able to iterate "
                                "through signals")

            # if checking Signal type (default) then
            # make sure container has signals only, quit iterating as soon as a
            # not-complying signal is found.
            if self._check_signal_type and \
               any(not isinstance(signal, Signal) for signal in signals):
                raise \
                    TypeError("All signals must be instances of Signal")

            # determine if signals are to be cloned.
            clone_signals = \
                self._clone_signals and len(self._receivers[block.id()]) > 1

            for receiver_data in self._receivers[block.id()]:
                if receiver_data.block.status.is_set(RunnerStatus.error):
                    self.logger.debug(
                        "Block '{}' has status 'error'. Not delivering "
                        "signals from '{}'...".format(
                            receiver_data.block.label(), block.label()))
                    continue
                elif receiver_data.block.status.is_set(RunnerStatus.warning):
                    self.logger.debug(
                        "Block '{}' has status 'warning'. Delivering signals"
                        " anyway from '{}...".format(
                            receiver_data.block.label(), block.label()))

                # We only send signals if the receiver's output matches
                # the output that these signals were notified on
                if receiver_data.output_id == output_id:
                    try:
                        signals_to_send = deepcopy(signals) \
                            if clone_signals else signals
                    except:
                        # if deepcopy fails, send original signals
                        signals_to_send = signals
                        self.logger.info("'deepcopy' operation failed while "
                                         "sending signals originating from "
                                         "block: {}".format(block.label()),
                                         exc_info=True)

                    self.logger.debug(
                        "Routing {} signals from {} to {}".format(
                            len(signals_to_send),
                            block.label(True),
                            receiver_data.block.label()))

                    if self._diagnostics:
                        self._diagnostic_manager.on_signal_delivery(
                            block.type(),
                            block.id(),
                            receiver_data.block.type(),
                            receiver_data.block.id(),
                            len(signals_to_send)
                        )

                    self.deliver_signals(receiver_data, signals_to_send)

        elif self.status.is_set(RunnerStatus.stopped):
            self.logger.warning("Block Router is stopped, discarding signal"
                                " notification from block: {}".
                                format(block.label()))
        elif self.status.is_set(RunnerStatus.stopping):
            self.logger.debug("Block Router is stopping, discarding signal"
                              " notification from block: {}".
                              format(block.label()))
        else:
            self.logger.warning("Block Router is not started, status is: {}, "
                                "discarding signal notification from block: {}"
                                .format(self.status, block.label()))
            raise BlockRouterNotStarted()

    def deliver_signals(self, block_receiver, signals):
        """ Overridable method to deliver signals to a block

        Other block routers can override this method to perform special actions
        when delivering signals to a block. It is recommended to call the
        block router's notify_signals_to_block method from this method though,
        as that takes care of handling different formats of blocks.

        Args:
            block_receiver (BlockReceiverData): The data about the block that
                will be receiving the signals
            signals (Iterable): The signals that the block will receive
        """
        self.notify_signals_to_block(block_receiver, signals)

    def notify_signals_to_block(self, block_receiver, signals):
        """ Call process signals on a block with the proper signature.

        Block developers are not required to include an input_id in their
        block's process_signals function definition. This method will
        """

        # Router subclasses end up calling this method when overriding
        # 'deliver_signals', so it is best to provide exception handling at
        # this level, Note: a given router, for example: ThreadPoolExecutor,
        # might not even offer a way to catch an exception, so catching
        # exceptions at this 'root' level, for ALL routers, makes sense
        try:
            # Check if block has defined the input_id in its process_signals
            if block_receiver.include_input_id:
                # Pass the block_receiver's input_id to the
                # process_signals method
                block_receiver.block.process_signals(
                    signals, block_receiver.input_id)
            else:
                # Only send the signals to the block, no input_id
                block_receiver.block.process_signals(signals)
        except:
            # TODO: define logic to set router back to normal
            self.status.add(RunnerStatus.error)
            self.logger.exception("{}.process_signals failed".
                                  format(block_receiver.block.label()))



from threading import RLock
from logging.handlers import RotatingFileHandler
from os import path, makedirs

MAX_FILE_SIZE = 1024 * 1024 * 15  # 15 MB


class NIOFileHandler(RotatingFileHandler):

    """ File handler for logging. Automatically rolls over at maxBytes.

    Instances of this class will be registered with the base Logger on an
    as-needed basis.

    """

    def __init__(self, dirname='.', filename='log_placeholder',
                 max_bytes=MAX_FILE_SIZE, backup_count=1):
        self.dirname = dirname
        self.filename = filename
        fname = self._construct_fname()
        super().__init__(fname, maxBytes=max_bytes,
                         backupCount=backup_count)

    def _construct_fname(self):
        """ Build the log-file name based on date and the name of logger.

        Args:
            name (str): the name of the logger to which this handler is
            registered

        Returns:
            (str): The basename for the log file. Successive integers will
            be appended as logs roll over.

        """
        # create the target directory if it doesn't exist
        if not path.isdir(self.dirname):
            makedirs(self.dirname)

        f = "{0}.log".format(self.filename)
        return path.join(self.dirname, f)


class NIOShareFileHandler(NIOFileHandler):

    """ NIO File Handler that caches open streams to the log file

    In Windows when more than one stream causes:

    PermissionError: [WinError 32] The process cannot access the file because
    it is being used by another process:

    """
    file_handlers = {}

    def __init__(self, dirname='.', filename='log_placeholder',
                 max_bytes=MAX_FILE_SIZE, backup_count=1):
        self._lock = RLock()
        super().__init__(dirname, filename, max_bytes, backup_count)

    def doRollover(self):
        """ Removes the cached handler before doing the doRollOver """
        with self._lock:
            self.__class__.file_handlers[self.baseFilename] = None
        super().doRollover()

    def _open(self):
        """ Open the file used for logging

        Caches an open stream to a file so the file can be rollover
        and renamed.
        """
        with self._lock:
            stream = self.__class__.file_handlers.get(self.baseFilename)
            if not stream:
                stream = super()._open()
                self.__class__.file_handlers[self.baseFilename] = stream
            return stream



from nio.modules.communication.publisher import Publisher
from nio.modules.communication.subscriber import Subscriber
from nio.modules.module import Module


class CommunicationModule(Module):

    def proxy_publisher_class(self, publisher_class):
        Publisher.proxy(publisher_class)

    def proxy_subscriber_class(self, subscriber_class):
        Subscriber.proxy(subscriber_class)

    def finalize(self):
        Publisher.unproxy()
        Subscriber.unproxy()
        super().finalize()

    def get_module_order(self):
        return 50

    def get_module_type(self):
        return "communication"



class Parameter(object):

    """ A typed command parameter. Base class for other parameters

    Examples of other command parameters are IntParameter, StringParameter, etc

    Args:
        type (Type): The type of data of this parameter.
        name (str): Name of the parameter. Used to look up arguments.
        title (str): More detailed description of the parameter
        default (str): default value to be used in
        allow_none (bool): does this parameter need a value?

    """

    def __init__(self, _type, name,
                 title=None, default=None, allow_none=False, **kwargs):
        self.type = _type
        self.name = name
        self.title = title or name
        self.default = default
        self.allow_none = allow_none
        self.kwargs = kwargs

    def get_description(self):
        """ Boilerplate description of the parameter.

        Args:
            None

        Returns:
            description (dict): The description.

        """
        return {
            "title": self.title,
            "default": self.default,
            "allow_none": self.allow_none
        }

    def deserialize(self, value, **kwargs):
        # Allow property kwargs to be overriden by call to serialize
        merged_kwargs = self.kwargs.copy()
        merged_kwargs.update(**kwargs)
        return self.type.deserialize(value, **merged_kwargs)



class Store(object):
    """ Global store that allows to save persisted data across instances
    """
    values = {}
    collections = {}


class Persistence(object):

    """ An implementation of the Persistence proxy that works in memory.

    This will not actually persist any values, but will keep them in
    memory to make sure that save/load commands work properly.

    In other words, this will "persist" its data only as long as the current
    environment is active. This is useful in unit-tests and in environments
    where persistence is not desired, but it is hardly useful in practice.

    Persisted data is intended to remain available for all instances
    """

    def load(self, id, collection=None, default=None):
        if collection is not None:
            return Store.collections.get(collection, {}).get(id, default)
        else:
            return Store.values.get(id, default)

    def load_collection(self, collection, default=None):
        return Store.collections.get(collection, default)

    def save(self, item, id, collection=None):
        if collection is not None:
            if collection not in Store.collections:
                Store.collections[collection] = {}
            Store.collections[collection][id] = item
        else:
            Store.values[id] = item

    def save_collection(self, items, collection):
        Store.collections[collection] = items

    def remove(self, id, collection=None):
        if collection is not None:
            if collection in Store.collections and \
                            id in Store.collections[collection]:
                del Store.collections[collection][id]
        else:
            if id in Store.values:
                del Store.values[id]

    def remove_collection(self, collection):
        if collection in Store.collections:
            del Store.collections[collection]

    @classmethod
    def configure(cls, context):
        # clear Persistence store before each test starts
        Store.values = {}
        Store.collections = {}



from nio.testing.modules.web.engine import TestingWebEngine
from nio.testing.modules.web.server import TestingWebServer
from nio.testing.modules.web.http import TestingRequest, TestingResponse
from nio.modules.web.module import WebModule


class TestingWebModule(WebModule):

    """ A version of the web module used for unit tests.

    This will proxy a simple WebEngine used for testing. Right now, this module
    implementation does nothing. A simple web server implementation is planned.
    """

    def initialize(self, context):
        super().initialize(context)
        self.proxy_web_engine_class(TestingWebEngine)
        self.proxy_web_server_class(TestingWebServer)
        self.proxy_request_class(TestingRequest)
        self.proxy_response_class(TestingResponse)

    def finalize(self):
        super().finalize()



"""
  NIO test support base class

"""
import logging.config
import multiprocessing
from unittest import TestCase

# Testing module implementations
from nio.modules.settings import Settings
from nio.testing.module_initializer import TestingModuleInitializer
from nio.testing.modules.communication.module \
    import TestingCommunicationModule
from nio.testing.modules.persistence.module \
    import TestingPersistenceModule
from nio.testing.modules.scheduler.module import TestingSchedulerModule
from nio.testing.modules.security.module import TestingSecurityModule
from nio.testing.modules.settings.module import TestingSettingsModule
from nio.testing.modules.web.module import TestingWebModule


class NIOTestCase(TestCase):

    """ Base Unit Test case class

    Allows to customize environment information when
    running NIO unit tests

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._module_initializer = None
        logging.config.dictConfig(self.get_logging_config())
        # make sure modules are tearDown regardless of test outcome
        self.addCleanup(self.tearDownModules)

        # maintain a mapping from module to module_name
        self._modules_mapping = {}

    def setupModules(self):
        """ Initializes the modules that will be active when running the test

        There are three important methods to override in a test when
        wanting to customize the modules behaviour. These are:
            - get_test_modules
            - get_module
            - get_context

        """
        self._module_initializer = TestingModuleInitializer(self)

        modules = self.get_test_modules()
        for module_name in modules:
            module = self.get_module(module_name)
            self._modules_mapping[module] = module_name
            self._module_initializer.register_module(module)

        # Perform a safe initialization in case a proxy never got cleaned up
        self._module_initializer.initialize(safe=True)

    def get_module(self, module_name):
        """ Returns a module class type given a module name

        Override this method to customize functionality in a module or
        simply when needing a module other than the default testing module.

        Args:
            module_name (str): module name

        Returns:
            module class type

        """
        known_modules = {
            'scheduler': TestingSchedulerModule,
            'persistence': TestingPersistenceModule,
            'security': TestingSecurityModule,
            'communication': TestingCommunicationModule,
            'settings': TestingSettingsModule,
            'web': TestingWebModule
        }
        if module_name not in known_modules:
            raise ValueError("{} is not a valid module".format(module_name))
        return known_modules.get(module_name)()

    def get_context(self, module_name, module):
        """ Provides the context to use when initializing the module

        Args:
            module_name (str): module name
            module (type): module class type

        Returns:
            module context

        """
        return getattr(module, "prepare_core_context")()

    def tearDownModules(self):
        # Perform a safe finalization in case anything wasn't proxied first
        if self._module_initializer:
            self._module_initializer.finalize(safe=True)
            # avoid a second call to finalize
            self._module_initializer = None

    def get_test_modules(self):
        """ Returns set of modules to load during test

        Override this method to customize which modules you want to load
        during a test
        """
        return {'settings', 'scheduler', 'persistence'}

    def set_settings(self):
        """ Allows test to set settings after 'settings' module is initialized

        A given test can set settings by making calls such as:
            Settings.set([section], [option], value)

        """
        pass

    def get_logging_config(self):
        """ Specifies the default logging configuration """
        return {
            "version": 1,
            "formatters": {
                "default": {
                    "format": ("[%(niotime)s] NIO [%(levelname)s] "
                               "[%(context)s] %(message)s")
                }
            },
            "filters": {
                "niofilter": {
                    "()": "nio.util.logging.filter.NIOFilter"
                }
            },
            "handlers": {
                "default": {
                    "level": "NOTSET",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",
                    "formatter": "default",
                    "filters": ["niofilter"]
                }
            },
            "root": {
                "handlers": ["default"]
            }
        }

    def setUp(self):
        super().setUp()
        try:
            multiprocessing.set_start_method('spawn')
        except Exception as e:
            # ignore this error
            if isinstance(e, RuntimeError) \
                    and "context has already been set" in str(e):
                pass
            else:
                print('Setting multiprocess start method, details: {0}'.
                      format(str(e)))

        self.setupModules()

    def tearDown(self):
        try:
            Settings.clear()
        except NotImplementedError:
            # can be triggered if test chooses not to have a Settings module
            pass
        self.tearDownModules()
        super().tearDown()

    def get_module_name(self, module):
        return self._modules_mapping.get(module, None)


class NIOTestCaseNoModules(NIOTestCase):

    """ NIO Test case that don't need to use modules

    """

    def get_test_modules(self):
        return set()

    def setupModules(self):
        pass  # Not using functionality modules

    def tearDownModules(self):
        pass  # Not using functionality modules



from nio.types.base import Type


class ListType(Type):

    @staticmethod
    def serialize(value, list_obj_type, **kwargs):
        """ Convert a value to a JSON serializable value """
        serialized_list = []
        for el in value:
            if issubclass(list_obj_type, Type):
                serialized_list.append(list_obj_type.serialize(el, **kwargs))
            else:
                # the element of the list is not valid
                raise TypeError('Element of list property if wrong type. '
                                'Element of type "{}" should be type "{}"'
                                .format(list_obj_type, type(el)))
        return serialized_list

    @staticmethod
    def deserialize(value, list_obj_type, **kwargs):
        """ Convert value to list """
        if not isinstance(value, list):
            raise TypeError("Unable to cast value to list: {}".format(value))
        list_to_assign = []
        for el in value:
            if issubclass(list_obj_type, Type):
                list_to_assign.append(list_obj_type.deserialize(el, **kwargs))
            else:
                raise TypeError(
                    "Unable to cast value to list: {}".format(value))
        return list_to_assign



"""

    API Security

    Provides utility methods to ensure access to protected resources

"""
import threading

from nio.modules.security import Authorizer, SecureTask
from nio.modules.security import User


def get_user():
    """ Provides current user

    In the chance that no current user is available, a 'Guest' user is returned
    """
    thread = threading.current_thread()
    return getattr(thread, "user", User())


def set_user(user):
    """ Sets current user
    """
    setattr(threading.current_thread(), "user", user)


def clear_user():
    """ Clears current user
    """
    delattr(threading.current_thread(), "user")


def has_access(resource, permission):
    """ Finds out if current user has permission to access given resource
    """
    return Authorizer.is_authorized(get_user(),
                                    SecureTask(resource, permission))


def ensure_access(resource, permission):
    """ Ensures that current user has permission to access given resource

    Raises:
        Unauthorized: if the user has no permission to access given resource
    """
    Authorizer.authorize(get_user(), SecureTask(resource, permission))



import functools
import re

from nio.modules.communication.topic import LEVEL_ALLOWED_CHARACTERS, \
    is_topic_type_valid


# defines level separator
LEVEL_SEPARATOR = re.escape(".")


def _add_regex_ending(regex):
    """ Utility function to add a regular expression ending

    Args:
        regex (str): Incoming regular expression

    Returns:
        Same regular expression plus ending
    """
    return regex + "\Z"


def _translate_to_regex(pattern):
    """ Translate a topic pattern to regular expression

    Algorithm:
        Split subscriber string around “.”
        for each substring (each substring defines a level)
            Replace any double “**” with zero or more allowable
                set of characters
            Replace any “*” single with a single allowable
                set of characters
            Escape any “*” characters when using it along with other characters
        Join together all subscriber topic string regexes

    Args:
        pattern (str): pattern to translate

    Returns:
        A regular expression that will match all valid publishers.
        Note that in order to simplify the level matching, this regex expects
        every level of the topic tree to end with a level separator (.).
        This includes the lowest level of the tree. In other words, make sure
        your publisher topic strings also end with a . before matching them to
        this regular expression.
    """
    sub_patterns = pattern.split('.')
    for i in range(len(sub_patterns)):
        if sub_patterns[i] == "**":
            # allow 0 or more levels
            sub_patterns[i] = \
                "({}+{})*".format(LEVEL_ALLOWED_CHARACTERS, LEVEL_SEPARATOR)
        elif sub_patterns[i] == "*":
            # allow 1 level
            sub_patterns[i] = \
                "{}+{}".format(LEVEL_ALLOWED_CHARACTERS, LEVEL_SEPARATOR)
        elif "*" in sub_patterns[i]:
            sub_patterns[i] = \
                "{}{}".format(sub_patterns[i].replace('*', re.escape('*')),
                              LEVEL_SEPARATOR)
        else:
            sub_patterns[i] = "{}{}".format(sub_patterns[i], LEVEL_SEPARATOR)

    return _add_regex_ending(''.join(sub_patterns))


@functools.lru_cache()
def _compile_pattern(pattern):
    """ Compiles a pattern returning match operations to evaluate against.

    Args:
        pattern (str): Pattern string

    Returns:
        list of match objects

    """
    expression = _translate_to_regex(pattern)
    return re.compile(expression).match


def matches(sub_topic, pub_topic):
    """ Finds out if there is a match between publisher and subscriber

    Assumes that both topics have been validated using 'is_topic_valid'
    Note: Validation is expected to take place within module's implementation.

    Args:
        sub_topic (str): Subscriber topic (can contain wildcards)
        pub_topic (str): Publisher topic

    Return:
        True if they match, False otherwise
    """
    if not is_topic_type_valid(pub_topic) or \
       not is_topic_type_valid(sub_topic):
        # no possible match with an invalid topic
        return False

    # Add a trailing dot to the publisher topic to match the entire level
    # in the regex
    pub_topic += "."

    # compile sub_topic pattern as a regex and match it
    compiled_match = _compile_pattern(sub_topic)
    if compiled_match(pub_topic) is not None:
        return True

    return False



from datetime import timedelta
from nio.types import Type


class TimeDeltaType(Type):

    @staticmethod
    def serialize(value, **kwargs):
        """ Convert a value to a JSON serializable value """
        if isinstance(value, timedelta) or isinstance(value, dict):
            return TimeDeltaType._timedelta_to_dict(value)
        else:
            return value

    @staticmethod
    def deserialize(value, **kwargs):
        """ Convert value to timedelta """
        if isinstance(value, dict):
            try:
                return timedelta(**value)
            except:
                raise TypeError(
                    "Unable to cast value to timedelta: {}".format(value))
        else:
            return value

    @staticmethod
    def _timedelta_to_dict(td):
        if isinstance(td, timedelta):
            return {'days': td.days,
                    'seconds': td.seconds,
                    'microseconds': td.microseconds}
        elif isinstance(td, dict):
            default_td = {'days': 0,
                          'seconds': 0,
                          'microseconds': 0}
            default_td.update(td)
            return default_td
        else:
            return {}



from nio.types.base import Type


class FloatType(Type):

    @staticmethod
    def serialize(value, **kwargs):
        """ Convert a value to a JSON serializable value """
        return value

    @staticmethod
    def deserialize(value, **kwargs):
        """ Convert value to float """
        try:
            return float(value)
        except:
            raise TypeError("Unable to cast value to float: {}".format(value))



from nio.modules.communication.matching import matches


class PubSubManager(object):

    """ A Communication manager to allow simple pub/sub in unit tests

    This works with the unit-test Publisher and Subscriber to allow data
    to be published and subscribed to in the same process.
    """

    publishers = {}
    subscribers = []

    @classmethod
    def add_publisher(cls, publisher):
        """ Add a publisher to this manager.

        This will register the topic of the publisher with any matching
        Subscribers so that they will get called when data is published on
        this publisher.
        """
        cls.publishers[publisher] = []
        for subscriber in cls.subscribers:
            if cls._matches(subscriber.topic, publisher.topic):
                cls.publishers[publisher].append(subscriber)

    @classmethod
    def remove_publisher(cls, publisher):
        """ De-register a publisher with this manager """
        if publisher in cls.publishers:
            del cls.publishers[publisher]

    @classmethod
    def add_subscriber(cls, subscriber):
        """ Add a subscriber to this manager and subscribe to relevant data.

        This will add this subscriber's callback to any registered publishers
        whose topic match.
        """
        cls.subscribers.append(subscriber)
        for publisher in cls.publishers.keys():
            if cls._matches(subscriber.topic, publisher.topic):
                cls.publishers[publisher].append(subscriber)

    @classmethod
    def remove_subscriber(cls, subscriber):
        """ De-register the subscriber with this manager """
        for publisher_callbacks in cls.publishers.values():
            if subscriber in publisher_callbacks:
                publisher_callbacks.remove(subscriber)
        cls.subscribers.remove(subscriber)

    @classmethod
    def send(cls, publisher, signals):
        """ Send data from a publisher to any subscribed callbacks """
        for subscriber in cls.publishers[publisher]:
            try:
                subscriber.handler(signals, publisher.topic)
            except TypeError:
                subscriber.handler(signals)

    @classmethod
    def _matches(cls, sub_topic, pub_topic):
        """ Use a simple matching algorithm imitating wildcards on a path

        Args:
            sub_topic (str): sub_topic to match can contain wildcards (*)
            pub_topic (str): publisher topic

        Returns:
            True if they match, False otherwise
        """
        return matches(sub_topic, pub_topic)



from nio import discoverable
from nio.block.context import BlockContext
from nio.command import command
from nio.command.holder import CommandHolder
from nio.properties import PropertyHolder, VersionProperty, \
    BoolProperty, ListProperty, StringProperty, Property, SelectProperty
from nio.router.context import RouterContext
from nio.util.logging import get_nio_logger
from nio.util.logging.levels import LogLevel
from nio.util.flags_enum import FlagsEnum
from nio.util.runner import Runner, RunnerStatus
from nio.util.threading import spawn


class BlockExecution(PropertyHolder):

    """ An object containing a block and its receivers

    Defines a single block execution within a potential execution graph
    A block execution is defined by a block name (name) and a set of block
    receivers where each receiver is identified by a block name.

    This information is parsed/used by the Router which is able then, based on
    the sending block, to forward signals to its receivers
    """
    id = StringProperty(title="Id")
    receivers = Property(title="Receivers")


class BlockMapping(PropertyHolder):

    """ An obejct containing a block and its alias in the service

    Allows a mapping of a given block based on another block
    This information is parsed/used internally by the core system
    """
    name = StringProperty(title="Name")
    mapping = StringProperty(title="Mapping")


@command('status', method="full_status")
@command('heartbeat')
@command('runproperties')
@command('start')
@command('stop')
@discoverable
class Service(PropertyHolder, CommandHolder, Runner):

    """The base class for services.

    Once a service is created, the service is correctly executed
    by calling 'configure' and 'start'
    """

    version = VersionProperty(version='1.0.0')
    type = StringProperty(title="Type", visible=False, readonly=True)
    id = StringProperty(title="Id", visible=False, allow_none=False)
    name = StringProperty(title="Name", allow_none=True)

    # indicates whether service is to be started when nio starts
    auto_start = BoolProperty(title="Auto Start", default=True)
    # indicates the logging level
    log_level = SelectProperty(LogLevel, title="Log Level", default="NOTSET")

    # properties defining the service execution
    execution = ListProperty(BlockExecution, title="Execution", default=[])
    mappings = ListProperty(BlockMapping, title="Mappings", default=[])

    # System Metadata that can be used by API clients, which is serialized
    # along with any other service properties
    sys_metadata = StringProperty(title="Metadata", visible=True, default="")

    def __init__(self, status_change_callback=None):
        """ Create a new service instance.

        Args:
            status_change_callback: method to call when status changes

        Take care of setting up instance variables in your service's
        constructor.
        """

        self.logger = get_nio_logger('service')
        super().__init__(status_change_callback=status_change_callback)

        # store service type so that it gets serialized
        self.type = self.__class__.__name__

        self._block_router = None
        self.mgmt_signal_handler = None
        self._blocks = {}
        self.mappings = []

        self._blocks_async_configure = None
        self._blocks_async_start = None
        self._blocks_async_stop = None

    def start(self):
        """Overrideable method to be called when the service starts.

        The service creator can assume at this point that the service's
        configuration is complete.

        If overriding, The service creator should call the start method
        on the parent, after which it can assume that block router and blocks
        are started
        """
        if self._block_router:
            self._block_router.do_start()

        if self._blocks_async_start:
            self._execute_on_blocks_async("do_start")
        else:
            for block in self._blocks.values():
                block.do_start()

    def stop(self):
        """Overrideable method to be called when the service stops.

        If overriding, The service creator should call the stop method
        on the parent, after which it can assume that block router and blocks
        are stopped
        """
        if self._blocks_async_stop:
            self._execute_on_blocks_async("do_stop")
        else:
            for block in self._blocks.values():
                block.do_stop()

        if self._block_router:
            self._block_router.do_stop()

    def _execute_on_blocks_async(self, method):
        """ Performs given method on all blocks in an async manner

        A join operation is performed for each spawned thread, that way we
        can assure that each block gets a chance to execute fully before
        leaving this method.

        Since created threads are not exposed to caller, and there is a
        potential for a spawned thread to be non-responsive, it is recommended
        that this or a method in the stack hierarchy leading to this method be
        spawned itself.

        Args:
            method (str): method to execute on blocks

        """
        threads = []
        for block in self._blocks.values():
            threads.append(spawn(getattr(block, method)))

        for thread in threads:
            thread.join()

    def configure(self, context):
        """Configure the service based on the context

        Args:
            context (ServiceContext): Specifies the information
                used to configure the service including service
                properties, block classes and properties for each
                block included in the service execution.
        """
        # populate service properties
        self.from_dict(context.properties, self.logger)
        # verify that service properties are valid
        self.validate()

        # reset logger after modules initialization
        # and properties setting
        self.logger = get_nio_logger(self.label())
        self.logger.setLevel(self.log_level())

        # instantiate block router
        self.logger.debug("Instantiating block router: {0}.{1}".
                          format(context.block_router_type.__module__,
                                 context.block_router_type.__name__))
        self.mgmt_signal_handler = context.mgmt_signal_handler
        self._blocks_async_configure = context.blocks_async_configure
        self._blocks_async_start = context.blocks_async_start
        self._blocks_async_stop = context.blocks_async_stop
        self._block_router = context.block_router_type()

        # create and configure blocks
        configure_threads = []
        for block_definition in context.blocks:
            block_context = self._create_block_context(
                block_definition['properties'],
                context)
            # create block instance
            block = block_definition['type']()
            # configure it
            if self._blocks_async_configure:
                # guarantee 'id' property is assigned to be able to reference
                # id() property down below
                block.id = block_context.properties["id"]
                configure_threads.append(
                    spawn(block.do_configure, block_context))
            else:
                block.do_configure(block_context)
            # register it
            self._blocks[block.id()] = block
        # if configuration was async, ensure they are all done
        if configure_threads:
            for thread in configure_threads:
                thread.join()

        # populate router context and configure block router
        router_context = RouterContext(self.execution(),
                                       self._blocks,
                                       context.router_settings,
                                       context.mgmt_signal_handler,
                                       context.instance_id,
                                       self.id(),
                                       self.name())
        self._block_router.do_configure(router_context)

    def _create_block_context(self, block_properties, service_context):
        """Populates block context to pass to the block's configure method"""
        return BlockContext(
            self._block_router,
            block_properties,
            service_context.properties.get('id'),
            service_context.properties.get('name', ""),
            self._create_commandable_url(service_context.properties,
                                         block_properties.get('id')),
            self.mgmt_signal_handler
        )

    def _create_commandable_url(self, service_properties, block_alias):
        """ Get the commandable url of a block given its alias """

        return '/services/{0}/{1}/'.format(
            service_properties.get('id', ''), block_alias)

    @classmethod
    def get_description(cls):
        """ Retrieves a service description based on properties and commands

        Returns:
            Service description
        """
        properties = super().get_description()
        commands = cls.get_command_description()
        return {'properties': properties,
                'commands': commands}

    @property
    def blocks(self):
        """ Allow access to service blocks """
        return self._blocks

    def heartbeat(self):
        """ Allows monitoring

        Returns:
            Service status
        """
        return self.status

    def runproperties(self):
        """ Returns service runtime properties """
        return self.to_dict()

    def full_status(self):
        """Returns service plus block statuses for each block in the service"""

        # build a general-like service status.
        service_and_blocks_status = FlagsEnum(RunnerStatus)
        # initialize it with the service status itself
        service_and_blocks_status.flags = self.status.flags
        # go trough its blocks and grab their warning and error statuses if any
        for block_id in self._blocks:
            if self._blocks[block_id].status.is_set(RunnerStatus.warning):
                service_and_blocks_status.add(RunnerStatus.warning)
            elif self._blocks[block_id].status.is_set(RunnerStatus.error):
                service_and_blocks_status.add(RunnerStatus.error)

        status = {"service": self.status.name,
                  "service_and_blocks": service_and_blocks_status.name,
                  "blocks": {}}
        # create a dict for all blocks using block label as key
        blocks = {}
        for block_id in self._blocks:
            block_status = {"status": self._blocks[block_id].status.name}
            if self._blocks[block_id].status.is_set(RunnerStatus.warning):
                block_status["warning"] = \
                    self._blocks[block_id]._messages[RunnerStatus.warning]
            if self._blocks[block_id].status.is_set(RunnerStatus.error):
                block_status["error"] = \
                    self._blocks[block_id]._messages[RunnerStatus.error]

            blocks[self._blocks[block_id].label()] = block_status

        status["blocks"] = blocks
        return status

    def label(self, include_id=False):
        """ Provides a label to a service based on name and id properties

        Args:
            include_id: whether id is to be included in label
        """
        if self.name():
            if include_id:
                return "{}-{}".format(self.name(), self.id())
            else:
                return self.name()
        return self.id()



from enum import Enum

from nio.util.flags_enum import FlagsEnum
from nio.util.logging import get_nio_logger


class RunnerStatus(Enum):
    """ Runner Status

    Status a runner can be at any time
    """
    created = 1
    configuring = 2
    configured = 3
    stopping = 4
    stopped = 5
    starting = 6
    started = 7
    warning = 8
    error = 9


class Runner(object):
    def __init__(self, *args, status_change_callback=None, **kwargs):
        """ Create a new runnable instance.

        Initializes runnable status

        """
        super().__init__(*args, **kwargs)
        self._status = FlagsEnum(RunnerStatus,
                                 status_change_callback=status_change_callback)
        # This can be overridden in subclasses if desired
        self.logger = get_nio_logger(self.__class__.__name__)
        self.status = RunnerStatus.created

    def configure(self, context):
        """Overrideable method to be called when the runnable is configured

        Args:
            context: specific information needed for runnable's configuration
        """
        pass  # pragma: no cover

    def start(self):
        """Overrideable method to be called when the runnable starts"""
        pass  # pragma: no cover

    def stop(self):
        """Overrideable method to be called when the runnable stops"""
        pass  # pragma: no cover

    @property
    def status(self):
        """ Provides component status """
        return self._status

    @status.setter
    def status(self, status):
        """ Status

        Possible values are based on RunnerStatus

        """
        self.logger.info("Setting status to {}".format(status.name))
        self.status.set(status)

    def do_configure(self, context):
        """ Entry point method to configure runnable.

        Ensures status is correctly set based on the outcome of the operation

        Args:
            context: specific information needed for runnable's configuration

        """
        self.status = RunnerStatus.configuring
        try:
            self.configure(context)
            self.status.replace(RunnerStatus.configuring,
                                RunnerStatus.configured)
        except Exception:
            self.logger.exception("Failed to configure")
            self.status.add(RunnerStatus.error)
            raise

    def do_start(self):
        """ Entry point method to start runnable.

        Ensures status is correctly set based on the outcome of the operation

        """
        self.status.replace(RunnerStatus.configured, RunnerStatus.starting)
        try:
            self.start()
            self.status.replace(RunnerStatus.starting, RunnerStatus.started)
        except Exception:
            self.logger.exception("Failed to start")
            self.status.add(RunnerStatus.error)
            raise

    def do_stop(self):
        """ Entry point method to stop runnable.

        Ensures status is correctly set based on the outcome of the operation

        """
        if self.status.is_set(RunnerStatus.stopped) or \
           self.status.is_set(RunnerStatus.stopping):
            self.logger.info("Already stopping or stopped")
            return

        self.status.replace(RunnerStatus.started, RunnerStatus.stopping)
        try:
            self.stop()
            self.status.replace(RunnerStatus.stopping, RunnerStatus.stopped)
        except Exception:
            self.logger.exception("Failed to stop")
            self.status.add(RunnerStatus.error)



import copy


class InvalidFlag(Exception):
    pass


class FlagsEnum(object):

    """Adds flag like functionality to an enum type

    For example:
        class Status(Enum):
            started = 1
            stopped = 2
            error_1 = 3
            error_2 = 4

        status = FlagsEnum(Status)
        status.add(Status.started)
        ....
        # an 'error_1' occurred
        status.add(Status.error_1)
        # Note that 'started' is still set
        ....
        # system recovered
        status.remove(Status.error_1)
        # Note: 'started' still remains, back to normal

    """

    def __init__(self, enum, default_flag=None,
                 status_change_callback=None):
        """ Create a new FlagsEnum instance.

        Args:
            enum (Enum): the enum from which flags are accepted
            default_flag (Enum): optional initial flag
            status_change_callback (callable): Method to call when
                detected a change in the flags value.

        Returns:
            None
        """
        super().__init__()

        self._enum = enum
        self._status_change_callback = status_change_callback
        self._flags = {}

        self.clear()
        if default_flag:
            self.add(default_flag)

    def _validate_flag(self, flag):
        """ Validates if given flag is within the range of the enum values

        Args:
            flag (value within the enum): the flag to check against

        Returns:
            None

        Raises:
            InvalidFlag
        """
        if flag not in self._enum:
            raise InvalidFlag('set, Invalid flag')

    def add(self, flag):
        """ Adds a flag value to the current set of flags

        Args:
            flag (value within the enum): the flag to add

        Returns:
            None

        Example:
            status = FlagsEnum(Status)
            status.add(Status.created)
        """
        self._validate_flag(flag)

        # make sure flag is not already set
        if not self._flags[flag.name]:

            # save old status to send along with changed status
            old_status = copy.deepcopy(self)
            self._flags[flag.name] = True
            if self._status_change_callback:
                self._status_change_callback(old_status, self)

    def replace(self, old_flag, new_flag):
        self._validate_flag(old_flag)
        self._validate_flag(new_flag)

        # save old status to send along with changed status
        old_status = copy.deepcopy(self)

        status_changed = False
        # remove old flag
        if self._flags[old_flag.name]:
            self._flags[old_flag.name] = False
            status_changed = True

        # add new flag
        if not self._flags[new_flag.name]:
            self._flags[new_flag.name] = True
            status_changed = True

        if status_changed and self._status_change_callback:
            self._status_change_callback(old_status, self)

    def remove(self, flag):
        """ Removes a flag value from the current set of flags

        Args:
            flag (value within the enum): the flag to remove

        Returns:
            None

        Example:
            status = FlagsEnum(Status)
            status.remove(Status.created)
        """
        self._validate_flag(flag)

        # make sure flag is set
        if self._flags[flag.name]:

            # save old status to send along with changed status
            old_status = copy.deepcopy(self)
            self._flags[flag.name] = False
            if self._status_change_callback:
                self._status_change_callback(old_status, self)

    def set(self, flag):
        """ Sets a flag, override any flags previously added

        Args:
            flag (value within the enum): the flag to set

        Returns:
            None

        Example:
            status = FlagsEnum(Status)
            # set status to be 'created'
            status.set(Status.created)
            # set status to be 'created, error'
            status.add(Status.error)
            # override previous status, set it to 'configured'
            status.set(Status.configured)
        """
        self._validate_flag(flag)

        # save old status to send along with changed status
        old_status = copy.deepcopy(self)
        change_occurred = False
        # set all flags but intended flag to False
        for key in self._flags.keys():
            if self._flags[key] and key != flag.name:
                self._flags[key] = False
                change_occurred = True

        # set intended flag if not already set
        if not self._flags[flag.name]:
            self._flags[flag.name] = True
            change_occurred = True

        if change_occurred and self._status_change_callback:
            self._status_change_callback(old_status, self)

    def is_set(self, flag):
        """ Checks if a flag is set

        Args:
            flag (value within the enum): the flag to check against

        Returns:
            True if flag is set, False otherwise

        """
        self._validate_flag(flag)
        return self._flags[flag.name]

    def clear(self):
        """ Resets all possible flag values

        Returns:
            None

        """
        for value in self._enum:
            self._flags[value.name] = False

    @property
    def flags(self):
        """ Provides flag values

        Returns:
            Flag values

        """
        return self._flags

    @flags.setter
    def flags(self, flags):
        """ Allows to set more than one flag at the same time

        Args:
            flags: flags to set

        Returns:
            None

        """

        # save old status to send along with changed status
        old_status = copy.deepcopy(self)
        change_occurred = False
        for key in flags.keys():
            if key in self.flags and self.flags[key] != flags[key]:
                change_occurred = True
                break

        self._flags = copy.deepcopy(flags)
        if change_occurred and self._status_change_callback:
            self._status_change_callback(old_status, self)

    @property
    def name(self):
        """ Provides a string with all flags currently set

        Returns:
            strings with all flag names currently set

        Example:
            status = FlagsEnum(Status)
            print(status.name)
        """
        statuses = ""
        for key in self._flags:
            if self._flags[key]:
                statuses += "{0}, ".format(key)
        return statuses[:-2]

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __eq__(self, rhs):
        """ Check if current settings matches the flag

        Args:
            rhs (flag or instance): right hand side part of equality
                setting to match to

        Returns:
            True or False

        """
        if isinstance(rhs, FlagsEnum):
            for key in self._flags:
                if key not in rhs.flags or rhs.flags[key] != self.flags[key]:
                    return False
        else:
            flag = rhs
            self._validate_flag(flag)
            if not self.is_set(flag):
                return False
            # make sure that all others are turned off
            for key in self._flags.keys():
                if key != flag.name and self._flags[key]:
                    return False

        return True

    def __getstate__(self):
        """ Make sure the object is able to be pickled.

        This method allows control over pickling, removes callback field
        which is not picklable.

        Returns:
            Fields to serialize

        """
        # make sure callback like attribute is present, since once it is
        # cloned once it will not be present any longer.
        if "_status_change_callback" in self.__dict__:
            # copy the dict since it will be changed
            odict = self.__dict__.copy()
            # remove callback entry
            del odict['_status_change_callback']
            return odict
        else:
            return self.__dict__



class AllowNoneViolation(Exception):
    """ Raised when a property contains a None value but is not allowed to """
    pass


class NoClassVersion(Exception):
    pass


class NoInstanceVersion(Exception):

    def __init__(self, class_version):
        super().__init__()
        self.class_version = class_version


class OlderThanMinVersion(Exception):

    def __init__(self, instance_version, min_version):
        super().__init__()
        self.instance_version = instance_version
        self.min_version = min_version


class InvalidEvaluationCall(Exception):
    pass




from nio.modules.proxy import ModuleProxy


class PublisherError(Exception):
    pass


class Publisher(ModuleProxy):

    """ An interface for a generic Publisher in the communication module """

    def __init__(self, topic=None, **kwargs):
        """ Create a new publisher instance.

        Args:
            context: Information that can optionally be provided specific to
                the communication implementation.

                It can be used to customize a publisher by overriding settings
                such as 'heartbeat_interval' and 'heartbeat_check_interval'.

            topic (str): Defines the kind of information to publish.

            kwargs: Key value argument pairs, used for backwards compatibility.
        """
        super().__init__(topic=topic, **kwargs)

    def open(self, on_connected=None, on_disconnected=None):
        """ Opens publishing channel

        Args:
            on_connected (callable): function receiving notification when
                connection is established
            on_disconnected (callable): function receiving notification when
                a disconnection occurs

        This method will setup the channel for sending, it is implementation
        specific, for example in a tcp like implementation it will open a port.

        """
        raise NotImplementedError()

    def is_connected(self):
        """ Determine if this publisher is connected and ready

        """
        raise NotImplementedError()

    def send(self, signals):
        """ Sends signals

        Args:
            signals: Signals to send
        """

        raise NotImplementedError()

    def close(self):
        """ Closes publisher.

        No further signals will be sent after a publisher is closed.

        """
        raise NotImplementedError()

    def is_closed(self):
        """ Finds out if publisher has been closed

        """
        raise NotImplementedError()



import json
import os
import pickle as unsafepickle

from safepickle import safepickle as pickle


def load_json(path):
    """ Loads a file in json format

    Args:
        path (str): path to file

    Returns:
        file contents as a dictionary
    """
    data = {}
    if os.path.isfile(path):
        with open(path, 'r') as f:
            data = json.loads(f.read())
    return data


def save_json(path, data):
    """ Saves a file in json format

    Args:
        path (str): path to file
        data (dict): data to save

    """
    with open(path, 'w+') as f:
        json.dump(data, f,
                  indent=4, separators=(',', ': '), sort_keys=True)


def load_pickle(path):
    """ Loads a file in pickle format

    Args:
        path (str): path to file

    Returns:
        file contents as a dictionary
    """
    data = {}
    if os.path.isfile(path):
        try:
            with open(path, 'r') as f:
                data = pickle.load(f)
        except:
            # TODO: this will be eventually removed once there is certainty
            # that all nio instances have been migrated to using safepickle
            with open(path, 'rb') as f:
                data = unsafepickle.load(f)
    return data


def save_pickle(path, data):
    """ Saves a file in pickle format

    Args:
        path (str): path to file
        data (dict): data to save

    """
    with open(path, 'w+') as f:
        pickle.dump(data, f)



from nio.modules.initializer import ModuleInitializer


class TestingModuleInitializer(ModuleInitializer):

    """ Allows to customize a module initializer for testing purposes """

    def __init__(self, test):
        super().__init__()
        # safe a test instance copy to delegate calls to it as needed
        self._test = test

    def _initialize_module(self, module, context, safe):
        """ Overrides module initialization

        Carries on with regular module initialization and in addition
        detects when testing module is initialized so that it can delegate
        the call to set settings on the test

        """
        super()._initialize_module(module, context, safe)

        # if it is the 'settings' module, allow test to set settings
        # before other modules get initialized
        module_name = self._test.get_module_name(module)
        if module_name == 'settings':
            self._test.set_settings()

    def get_context(self, module):
        """ Overrides initializer's get_context so that call can be routed
        to actual test
        """
        module_name = self._test.get_module_name(module)

        # make sure to call get_context on the test since it is a method
        # every test can redefine
        return self._test.get_context(module_name, module)



from nio.testing.modules.communication.publisher import Publisher
from nio.testing.modules.communication.subscriber import Subscriber
from nio.modules.communication.module import CommunicationModule


class TestingCommunicationModule(CommunicationModule):

    """ A version of the communication module used for unit tests.

    This will proxy the Publisher and Subscriber classes that are also
    used for unit tests. These perform the pub/sub opertaions in memory.
    """

    def initialize(self, context):
        super().initialize(context)
        self.proxy_publisher_class(Publisher)
        self.proxy_subscriber_class(Subscriber)

    def finalize(self):
        super().finalize()



from nio.types.base import Type
from nio.properties.holder import PropertyHolder


class ObjectType(Type):

    @staticmethod
    def serialize(value, **kwargs):
        """ Convert a value to a JSON serializable value """
        if value is not None:
            if isinstance(value, dict):
                # Good to go if it's already a dict
                return value
            else:
                # but to_dict if we're serialzing the PropertyHolder
                return value.to_dict()

    @staticmethod
    def deserialize(value, **kwargs):
        """ Convert value to object """
        try:
            if isinstance(value, dict):
                # Object types supports deserializing dictionaries
                sub_instance = kwargs["obj_type"]()
                # First load the value from the dict into the property holder
                sub_instance.from_dict(value)
                # Then make sure the values are valid
                sub_instance.validate_dict(value)
                return sub_instance
            elif isinstance(value, PropertyHolder):
                # And object types also work when value is already a holder
                return value
            else:
                raise Exception
        except:
            raise TypeError("Unable to cast value to object: {}".format(value))



class HTTPError(Exception):

    def __init__(self, status=500, message=None):
        """ Create a new HTTPError instance.

        Args:
            status: http error status
            message: http optional error message
        """
        super().__init__()
        self.status = status
        self.message = message


class HTTPNotImplemented(HTTPError):

    """Exception raised for not implemented features."""

    def __init__(self):
        """ Create a new HTTPNotImplemented instance.
        """
        super().__init__(501, "Not Implemented")


class HTTPUnauthorized(HTTPError):

    """Exception raised for not authorized."""

    def __init__(self):
        """ Create a new HTTPUnauthorized instance.
        """
        super().__init__(401, "Unauthorized")


class HTTPNotFound(HTTPError):

    """Exception raised for not found errors."""

    def __init__(self):
        """ Create a new HTTPNotFound instance.
        """
        super().__init__(404, "Not Found")



from inspect import isclass


def ensure_is(obj, types, fn, msg=None):
    """ Make sure an object is of a given type

    General method for checking an object against a
    list of types. If the object does not match one of the
    types in 'types', then 'fn' (parameterized by 'obj') is
    called. If 'fn' is an Exception class, then an instance
    of that Exception class is raised.

    Args:
        obj: The object under consideration.
        types (list): A list of types.
        fn (callable): This is called when type(obj) not in types
        msg (string): Optional message for the case when 'fn' is
            as Exception

    Returns:
        obj: Either the original object (if the types match) or
            the result of 'fn(obj)'

    """
    for t in types:
        if isinstance(obj, t):
            return obj

    if isclass(fn) and issubclass(fn, Exception):
        raise fn() if msg is None else fn(msg)
    return fn(obj)


def ensure_list(obj, fn=lambda x: [x], msg=None):
    """ Make sure an object is a list.

    Less general version for the common case.
    A custom callable may be included, but the default
    parameters are convenient for checking whether an
    object is a list, wrapping it in a list if necessary.

    Args:
        obj: The object under consideration
        fn (callable): This is called when type(obj) not in types
        msg (string): Optional message for the case when 'fn' is
            as Exception

    Returns:
        obj: Either the original object or the result of
            'fn(obj)'

    """
    return ensure_is(obj, types=[list], fn=fn, msg=msg)



from nio.signal.base import Signal


class ManagementSignal(Signal):

    """ A special type of signal for management of n.io components.

    A management signal is a type of signal that n.io binaries and other
    plugins will sometimes look for. Any information that needs to be passed
    between n.io components for internal use will be in the form of a
    ManagementSignal. This interface is also available to block developers
    to notify information.

    This signal does not have any other attributes or functionality forced
    on it, it is just a placeholder super class to be used.
    """
    pass



class BackoffStrategy(object):

    def __init__(self, logger=None, max_retry=0, multiplier=1,
                 indefinite=False, **kwargs):
        """ Create an instance of a backoff strategy

        Args:
            logger (Logger): The logger to use when retrying
            max_retry (int): The maximum number of retries to try before
                stopping any increasing delay. Depending on the value of the
                indefinite parameter, retrying may give up at this point.
            multiplier (float): The multiplier to apply when sleeping. By
                default, this strategy will sleep a number of seconds equal to
                the current retry number. This argument allows you to alter
                that. For example, by default the 5th retry will sleep for 5
                seconds. Passing a multiplier of 1.5 will cause the 5th retry
                to sleep for 7.5 seconds instead.
            indefinite (bool): Whether to continue on retrying indefinitely
                once the max duration is reached
        """
        super().__init__()
        self.logger = logger
        self.max_retry = max_retry
        self.multiplier = multiplier
        self.indefinite = indefinite
        self.retry_num = 0

    def request_failed(self, exc):
        """ Called when a retryable request fails """
        self.retry_num += 1

    def request_succeeded(self):
        """ Called when a retryable request succeeds """
        self.retry_num = 0

    def should_retry(self):
        """ Determine whether or not we should retry this time.

        This function is responsible for determining whether or not a retry
        should take place; the wait_for_retry function will do the actual
        waiting for the retry.

        This method has a default implementation that will check if the max
        number of retries has taken place or if the indefinite retry
        setting has been checked. In your strategy, if you wish to utilize
        this behavior, you should only need to override the wait_for_retry
        function.

        Returns:
            bool: True if the retry should execute. False if we are "done"
                retrying
        """
        self.logger.info("Executing retry number {}, max retries is {}".format(
            self.retry_num, self.max_retry))
        if self.max_retry >= 0 and self.retry_num > self.max_retry:
            # There is a configured max number of retries and we are
            # officially past it, see if we continue or not
            if not self.indefinite:
                self.logger.warning(
                    "Done retrying - we have exceeded the maximum number "
                    "of retries ({})".format(self.max_retry))
                return False
        # If we got here either we are retrying indefinitely or we haven't hit
        # our max yet, let's retry
        return True

    def wait_for_retry(self):
        """ Do any waiting before the next retry is actually executed.

        The strategy implementation will want to override this method to
        define how much time should elapse between retries.
        """
        pass

    def use_logger(self, logger):
        """Use a logger instance in this backoff strategy.

        This helper function is useful for blocks or mixins to call so that the
        backoff strategy can log from that block. This gets called
        automatically by the Retry mixin but you are free to make another call
        to use a different logger.

        Args:
            logger (Logger): A logger that the backoff strategy can use
        """
        self.logger = logger



""" The base class for creating nio Blocks.

A block contains modular functionality to be used inside of Services. To create
a custom block, extend this Block class and override the appropriate methods.
"""
from collections import defaultdict

from nio.block.context import BlockContext
from nio.block.terminals import Terminal, TerminalType, input, output, \
    DEFAULT_TERMINAL
from nio.command import command
from nio.command.holder import CommandHolder
from nio.properties import PropertyHolder, StringProperty, \
    VersionProperty, SelectProperty
from nio.router.base import BlockRouter
from nio.signal.base import Signal
from nio.signal.status import BlockStatusSignal
from nio.util.logging import get_nio_logger
from nio.util.logging.levels import LogLevel
from nio.util.runner import Runner, RunnerStatus


@command('properties')
class Base(PropertyHolder, CommandHolder, Runner):

    """The base class for blocks to inherit from."""

    version = VersionProperty(version='0.0.0', advanced=True)
    type = StringProperty(title="Type", visible=False, readonly=True)
    id = StringProperty(title="Id", visible=False, allow_none=False)
    name = StringProperty(title="Name", visible=False, allow_none=True)
    log_level = SelectProperty(enum=LogLevel,
                               title="Log Level", default="NOTSET", advanced=True)

    def __init__(self, status_change_callback=None):
        """ Create a new block instance.

        Take care of setting up instance variables in your block's constructor.
        Note that the properties of the block are not available when the block
        is created. Those will be available when the block is configured.

        It is normally meaningless to pass variables to the constructor of the
        block. Any data the block requires will be passed through the
        BlockContext when the block is configured.
        """

        # We will replace the block's logger with its own name once we learn
        # what that name is during configure()
        self.logger = get_nio_logger('default')

        super().__init__(status_change_callback=status_change_callback)

        # store block type so that it gets serialized
        self.type = self.__class__.__name__

        self._block_router = None
        self._service_id = None
        self._service_name = None
        self._mgmt_signal_handler = None
        self._default_input = Terminal.get_default_terminal_on_class(
            self.__class__, TerminalType.input)
        self._default_output = Terminal.get_default_terminal_on_class(
            self.__class__, TerminalType.output)
        self._messages = defaultdict(str)

    def configure(self, context):
        """Overrideable method to be called when the block configures.

        The block creator should call the configure method on the parent,
        after which it can assume that any parent configuration options present
        on the block are loaded in as class variables. They can also assume
        that all functional modules in the service process are loaded and
        started.

        Args:
            context (BlockContext): The context to use to configure the block.

        Raises:
            TypeError: If the specified router is not a BlockRouter
        """
        if not isinstance(context, BlockContext):
            raise TypeError("Block must be configured with a BlockContext")
        # Ensure it is a BlockRouter so we can safely notify
        if not isinstance(context.block_router, BlockRouter):
            raise TypeError("Block's router must be instance of BlockRouter")

        self._block_router = context.block_router

        # load the configuration as class variables
        self.from_dict(context.properties, self.logger)
        # verify that block properties are valid
        self.validate()

        self.logger = get_nio_logger(self.label())
        self.logger.setLevel(self.log_level())
        self._service_id = context.service_id
        self._service_name = context.service_name
        self._mgmt_signal_handler = context.mgmt_signal_handler

    def start(self):
        """Overrideable method to be called when the block starts.

        The block creator can assume at this point that the block's
        initialization is complete and that the service and block router
        are in "starting" state.
        """
        pass  # pragma: no cover

    def stop(self):
        """Overrideable method to be called when the block stops.

        The block creator can assume at this point that the service and block
        router are in "stopping" state. All modules are still available for use
        in the service process.
        """
        pass  # pragma: no cover

    def set_status(self, status, message="", replace_existing=False):
        """ Set this block's status and notify the management channel.

        This is a method that a block developer can use to set their block's
        status. A block in warning or error status may be noted in the
        system designer or in monitoring tools.

        In general a block in warning status has a problematic condition but
        may be able to recover on its own (e.g., retrying a connection). A
        block in error status is in a state that will require a human to
        resolve the error; either by fixing the error condition or restarting
        the service.

        Args:
            status (str or RunnerStatus): A nio.util.runner.RunnerStatus to
                set the block's statut to. Optionally takes a string if the
                string is one of (error, warning, ok). Specifying 'error' or
                'warning' will set those status flags on the block. Specifying
                any non-error status as a status will clear both any error or
                warning status flags on the block.
            message (str): An optional string message to include in the
                notification of the status change.
            replace_existing (bool): Whether or not to replace the existing
                block status when setting the new status. If False, the new
                status will be added to the block if it is a warning or error
                status. Any non-error/warning status will clear any error or
                warning status that was on the block before.

        Returns:
            RunnerStatus: The current block's status after setting.

        Raises:
            ValueError: If a string other than "ok", "warning", or "error" is
                provided as the status
            TypeError: If an invalid status type or message type is specified
        """
        if isinstance(status, str):
            if status.lower() == 'ok':
                # An ok status means we want the first non-error status that
                # is on the block right now.
                # Default to a created status if the block has no statues left
                status = RunnerStatus.created
                for flag, flag_set in self.status.flags.items():
                    if flag in ('warning', 'error'):
                        continue
                    if flag_set:
                        status = RunnerStatus[flag]
                        break
            elif status.lower() == 'warn' or status.lower() == 'warning':
                status = RunnerStatus.warning
            elif status.lower() == 'error':
                status = RunnerStatus.error
            else:
                raise ValueError("Only 'ok', 'warning', or 'error' are "
                                 "supported status strings")
        elif not isinstance(status, RunnerStatus):
            raise TypeError(
                "Block status can only be set to string or RunnerStatus")

        if not isinstance(message, str):
            raise TypeError("Only string based status messages are allowed")

        # Clear old error/warn status if a non-error/warn status is passed
        if status != RunnerStatus.error:
            self.status.remove(RunnerStatus.error)
            if RunnerStatus.error in self._messages:
                del self._messages[RunnerStatus.error]
        else:
            self._messages[RunnerStatus.error] = message

        if status != RunnerStatus.warning:
            self.status.remove(RunnerStatus.warning)
            if RunnerStatus.warning in self._messages:
                del self._messages[RunnerStatus.warning]
        else:
            self._messages[RunnerStatus.warning] = message

        if replace_existing:
            self.status.set(status)
        else:
            self.status.add(status)

        # Notify the new status to the management channel for other services
        # to handle
        signal = BlockStatusSignal(status, message=message)
        self.notify_management_signal(signal)
        return self.status

    def notify_signals(self, signals, output_id=None):
        """Notify signals to router.

        This is the method the block should call whenever it would like
        to "output" signals for the router to send downstream.

        Args:
            signals (list): A list of signals to notify to the router
            output_id: The identifier of the output terminal to notify the
                signals on

        The signals argument is handled as follows:
            - a dictionary is not allowed
            - if a single signal is notified not as an iterable, it will get
            wrapped inside a list before forwarding to block router.

        Raises:
            TypeError: when signals are not instances of class Signal
        """

        if isinstance(signals, dict):
            raise TypeError("Signals cannot be a dictionary")

        # if a single Signal is being notified, make it a list
        if isinstance(signals, Signal):
            signals = [signals]

        self._block_router.notify_signals(self, signals, output_id)

    def notify_management_signal(self, signal):
        """Notify a management signal to router.

        Args:
            signal: signal to notify

        This is a special type of signal notification that does not actually
        propagate signals in the service. Instead, it is used to communicate
        some information to the block router about the block. For example,
        the block can report itself in an error state and thus prevent other
        signals from being delivered to it.
        """
        if self._mgmt_signal_handler:
            if isinstance(signal, BlockStatusSignal):
                # add service and block info
                signal.service_id = self._service_id
                signal.service_name = self._service_name
                signal.block_id = self.id()
                signal.block_name = self.name()
            self._mgmt_signal_handler(signal)

    def process_signals(self, signals, input_id=DEFAULT_TERMINAL):
        """Overrideable method to be called when signals are delivered.

        This method will be called by the block router whenever signals
        are sent to the block. The method should not return the modified
        signals, but rather call `notify_signals` so that the router
        can route them properly.

        Args:
            signals (list): A list of signals to be processed by the block
            input_id: The identifier of the input terminal the signals are
                being delivered to
        """
        pass  # pragma: no cover

    @classmethod
    def get_description(cls):
        """ Get a dictionary description of this block.

        Returns:
            dict: A dictionary containing the blocks properties and commands
        """
        properties = super().get_description()
        commands = cls.get_command_description()

        return {
            'properties': properties,
            'commands': commands
        }

    def properties(self):
        """ Returns block runtime properties """
        return self.to_dict()

    @classmethod
    def inputs(cls):
        """ A list of the block's input terminals """
        return Terminal.get_terminals_on_class(cls, TerminalType.input)

    @classmethod
    def outputs(cls):
        """ A list of the block's output terminals """
        return Terminal.get_terminals_on_class(cls, TerminalType.output)

    def is_input_valid(self, input_id):
        """ Find out if input is valid

        Args:
            input_id: input identifier

        Returns:
            bool: True if the input ID exists on this block
        """
        return input_id in [i.id for i in self.__class__.inputs()]

    def is_output_valid(self, output_id):
        """ Find out if output is valid

        Args:
            output_id: output identifier

        Returns:
            bool: True if the output ID exists on this block
        """
        return output_id in [o.id for o in self.__class__.outputs()]

    def label(self, include_id=False):
        """ Provides a label to a block based on name and id properties

        Args:
            include_id: whether id is to be included in label
        """
        if self.name():
            if include_id:
                return "{}-{}".format(self.name(), self.id())
            else:
                return self.name()
        return self.id()


@input(DEFAULT_TERMINAL, default=True, label="default")
@output(DEFAULT_TERMINAL, default=True, label="default")
class Block(Base):
    pass



import re

from nio.properties.exceptions import InvalidEvaluationCall
from nio.properties.util.parser import Parser
from nio.signal.base import Signal


class TemporarySignal(Signal):

    def __getattribute__(self, name):
        raise InvalidEvaluationCall


class Evaluator:

    """ Evaluate python expressions against a Signal.

    Class for transforming NIO's dynamic signal access mini-language
    into valid Python.

    Creates a new parser object each time 'evaluate' is called (i.e. for each
    incoming signal).

    Args:
        expression (str): The string or expression to be interpolated or
            evaluated. If expression is not a string then the raw expression
            is returned when evaluated.

    Raises:
        Exception: Raise any python exception during evaluation

    """
    delimiter = re.compile(r'(?<!\\)({{|}})|(\s)')
    expression_cache = {}

    def __init__(self, expression):
        self.expression = expression

    def evaluate(self, signal=None):
        if not isinstance(self.expression, str):
            return self.expression
        cache_key = (self.expression)
        parsed = self.__class__.expression_cache.get(cache_key, None)
        if parsed is None:
            # Only parse the expression if we haven't already done it.
            tokens = self.tokenize(self.expression)
            parser = Parser()
            parsed = parser.parse(tokens)
            self.__class__.expression_cache[cache_key] = parsed
        try:
            result = self._eval(signal, parsed)
            result_len = len(result)
            if result_len > 1:
                result = ''.join([str(o) for o in result])
            elif result_len == 1:
                result = result[0]
            else:
                result = ''
        except:
            raise

        return result

    def _eval(self, signal, parsed):
        result = []
        for item in parsed:
            if hasattr(item, '__call__'):
                if signal is None:
                    # Use a temporary signal that raises InvalidEvaluationCall
                    signal = TemporarySignal()
                # Evaluate the expression against the signal
                item = item(signal, self)
                if isinstance(item, TemporarySignal):
                    # This is to catch evaluating the expression "{{ $ }}"
                    # when evaluated without a Signal
                    raise InvalidEvaluationCall
            result.append(item)
        return result

    def tokenize(self, expression):
        """ Pad the delimiters with whitespace and split the expression. """
        tokens = self.delimiter.split(expression)

        # the split includes a bunch of None's and empty strings...
        return [t for t in tokens if t]



from nio.command.params.base import Parameter
from nio.types import ListType


class ListParameter(Parameter):

    def __init__(self, list_obj_type, name, **kwargs):
        kwargs['list_obj_type'] = list_obj_type
        super().__init__(ListType, name, **kwargs)



class Entity(object):

    def __init__(self, data=None):
        if data is None:
            data = dict()

        if not isinstance(data, dict):
            raise TypeError("Entity data must be a dictionary")

        self.data = data



from nio.modules.proxy import ModuleProxy


class WebServer(ModuleProxy):

    """ A Web Server class

    A web server is created by a web engine and can be started, stopped, and
    maintains a list of handlers (WebHandler) that it can receive requests
    on.

    This will likely be implemented in the web module implementation.

    """

    @property
    def host(self):
        """ Provides server host """
        return None

    @property
    def port(self):
        """ Provides server port """
        return None

    def add_handler(self, handler):
        """ Adds a web handler

        Args:
            handler (WebHandler): Web handler

        """
        raise NotImplementedError()

    def start(self, configuration):
        """ Starts web server

        Args:
            configuration: Server configuration

        """
        raise NotImplementedError()

    def stop(self):
        """ Stops web server

        """
        raise NotImplementedError()



from nio.command.params.base import Parameter
from nio.types import StringType


class StringParameter(Parameter):

    def __init__(self, name, **kwargs):
        super().__init__(StringType, name, **kwargs)
